import numpy as np

from . import fortran as modfort 

from .base import Model

from ..fitpars import Par, Parameters

#from numba import prange,njit,jitclass,int32,float32 # jitclass doesn't work with abstract methods
from numba import prange,njit,int32,float32 # jitclass doesn't work with abstract methods

# IDEA: make a Parameter class, which supports bounds etc...
# this will help to get rid of "__npar__"
# Drawback: need to re-define all basic functions and operators
# for Parameter class
#https://www.python-course.eu/python3_magic_methods.php
         
class PolynomialModel1D(Model):
    """
    1D polynomial model 
    """
    def __init__(self,maxpower):
        self.__params__ = Parameters(maxpower)

    def __func__(self,params,x):
        p = params.get_values()
        return np.polyval(p,x)
            
def poly2d(p,x,y,powers):
    n = len(p)
    sum = 0.0
    for i in range(n):
        sum += p[i]*x**powers[i][0]*y**powers[i][1]
    return sum
fast_poly2d = njit(poly2d)

def generate_powers(maxpower):
    powers = []
    for pow_x in range(maxpower+1):
        for pow_y in range(maxpower+1-pow_x):
            powers.append([pow_x,pow_y])
    return powers

def generate_powers_layer_3d(total_power):
    """
    Generate layer of powers for 3d polynomial for a given total power (unsymmetrized)
    """
    powers = []
    for j_plus_k in range(total_power,-1,-1):
        i = total_power - j_plus_k
        for j in range(j_plus_k+1):
            k = j_plus_k - j
            powers.append([i,j,k])
    return powers

def generate_powers_layer_3d_symm(total_power):
    """
    Generate layer of powers for 3d polynomial for a given total power (symmetrized i<=j)
    """
    powers = []
    for k in range(total_power+1):
        i_plus_j = total_power-k
        for i in range(i_plus_j+1):
            j = i_plus_j - i
            if i<=j: powers.append([i,j,k])
    return powers
    
def generate_powers_3d_rsymm(maxpower):
    """
    Generate powers for polynomial symmetrized by first 2 coordinates (r1,r2)
    """
    powers = []
    for pow_alpha in range(maxpower+1): # power of alpha
        for pow_r1_plus_r2 in range(maxpower-pow_alpha+1): # sum of r1 and r2 powers
            for pow_r1 in range(pow_r1_plus_r2+1): # power of r1
                pow_r2 = pow_r1_plus_r2-pow_r1
                if pow_r1<pow_r2: continue
                powers.append([pow_r1,pow_r2,pow_alpha])
    return powers

def poly3d(p,x,y,z,powers):
    n = len(p)
    sum = 0.0
    for i in range(n):
        sum += p[i]*x**powers[i][0]*y**powers[i][1]*z**powers[i][2]
    return sum
fast_poly3d = njit(poly2d)
    
def poly3d_rsymm(p,r1,r2,alpha,powers):
    """
    3D polynomial symmetrized by first 2 coordinates (r1,r2)
    """
    n = len(p)
    sum = 0.0
    for i in range(n):
        pow_r1 = powers[i][0]
        pow_r2 = powers[i][1]
        pow_alpha = powers[i][2]
        sum += p[i]*(r1**pow_r1*r2**pow_r2+r1**pow_r2*r2**pow_r1)*alpha**pow_alpha
    return sum
fast_poly3d_rsymm = njit(poly3d_rsymm)
    
class PolynomialModel2D(Model):
    """
    2D polynomial model
    """
    def __init__(self,maxpower,fast=False):
        self.__powers__ = np.array(generate_powers(maxpower))
        self.__params__ = Parameters(len(self.__powers__),
            names=['%d_%d'%tuple(_) for _ in self.__powers__],
            group='linear')
        self.__fast__ = fast
                    
    def __func__(self,params,x,y):
        p = params.get_values()
        if self.__fast__:
            res = fast_poly2d(p,x,y,self.__powers__)
        else:
            res = poly2d(p,x,y,self.__powers__)
        return res

# ========================================        
# ========================================        
# MODELS FOR FITTING THE 666 ISOTOPOLOGUE        
# ========================================        
# ========================================        
        
class DBOCModel_1(Model):
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
            [0,     0], # 0
            
            [1,     0], # 1
            [0,     1], # 1
            
            [1,     1], # 2
            [0,     2], # 2
            [2,     0], # 2
            
            [3,     0], # 3
            [2,     1], # 3
            [1,     2], # 3
            [0,     3], # 3
            
            [4,     0], # 4
            [3,     1], # 4
            [2,     2], # 4
            [1,     3], # 4
            [0,     4], # 4            
        ])
        self.__params__ = Parameters(len(self.__powers__),
            names=['%d_%d'%tuple(_) for _ in self.__powers__])
            
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        p = params.get_values()
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
        if self.__fast__:
            res = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            res = poly2d(p,r_part,alpha_part,self.__powers__)
        return res
        
class DBOCModel_2(Model): # GOOD MODEL TO FIT DBOC(6,6) FOR 666 ISOTOPOLOGUE; DON'T CHANGE
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__components__ = {}
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
        #   r_symm  alpha
            [0,     0], # 0   
            
            [1,     0], # 1   
            [0,     1], # 1   
            
            [1,     1], # 2   
            [0,     2], # 2   
            [2,     0], # 2   
        ])
                
        # polynomial parameters, active by default, no bounds
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,flag=True)
        
        # gaussian parameters, with bounds
        self.__params__.append(group='gaussian', pars=[
            Par(name='r_gauss', value=2.01, min=2.0, max=4.0,  flag=True),
            Par(name='a_gauss', value=9.99964, min=2.0, max=10.0, flag=True),
        ])
        
        # structure parameters, with bounds
        weight = 0.0
        self.__params__.append(group='structure', pars=[
            Par(name='r0_struct',value=2.95988, min=1.0, max=5.0,  weight=weight, flag=True),
            Par(name='a_struct', value=2.1391,  min=1,   max=50.0, weight=weight, flag=True),
            Par(name='h_struct', value=7.0, min=-np.inf, max=np.inf, weight=weight, flag=True),
            Par(name='apoly0_struct', value=3.01235,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='apoly1_struct', value=1.06193,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='apoly2_struct', value=0.269482, min=-np.inf, weight=weight, max=np.inf, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=0.0,  min=-30.0,   max=30.0,  flag=False),
        ])
                
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        r_gauss = params['r_gauss'].get_value()
        a_gauss = params['a_gauss'].get_value()
        gauss_r1 = np.exp(-a_gauss*(r1_au-re_au)**2)
        gauss_r2 = np.exp(-a_gauss*(r2_au-re_au)**2)
        gauss = gauss_r1*gauss_r2
        
        # get structural parameters
        r0_struct = params['r0_struct'].get_value()
        a_struct = params['a_struct'].get_value()
        h_struct = params['h_struct'].get_value()
        apoly0_struct = params['apoly0_struct'].get_value()
        apoly1_struct = params['apoly1_struct'].get_value()
        apoly2_struct = params['apoly2_struct'].get_value()
        struct_r1 = 0.5*(np.tanh(a_struct*(r1_au-r0_struct)/r0_struct)+1)
        struct_r2 = 0.5*(np.tanh(a_struct*(r2_au-r0_struct)/r0_struct)+1)
        struct = h_struct*struct_r1*struct_r2
        struct *= (apoly0_struct+apoly1_struct*alpha_part+apoly2_struct*alpha_part**2)
        
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # get final result
        res = poly*gauss+struct+global_shift
        
        # set components
        self.__components__['gauss'] = gauss
        self.__components__['poly'] = poly
        self.__components__['polygauss'] = poly*gauss
        self.__components__['struct'] = struct 
        self.__components__['globshift'] = global_shift
        
        return res
      
class DBOCModel_test(Model): # SANDBOX MODEL FOR TESTING PURPOSES
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__components__ = {}
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
            [0,     0], # 0   
            
            [1,     0], # 1   
            [0,     1], # 1   
            
            [1,     1], # 2   
            [0,     2], # 2   
            [2,     0], # 2   
            
            [3,     0], # 3
            [2,     1], # 3
            [1,     2], # 3
            [0,     3], # 3
            
            [4,     0], # 4
            [3,     1], # 4
            [2,     2], # 4
            [1,     3], # 4
            [0,     4], # 4            
        ])
                
        # polynomial parameters, active by default, no bounds
        weight = 0.0
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,weight=weight,flag=True)
        
        # gaussian parameters, with bounds
        weight = 5.0
        self.__params__.append(group='gaussian', pars=[
            Par(name='r_gauss', value=2.5, min=2.0, max=3.0, weight=weight, flag=True),
            Par(name='a_gauss', value=3.5, min=2.0, max=10.0, weight=weight, flag=True),
        ])
        
        # structure parameters, with bounds
        weight = 0.0
        self.__params__.append(group='structure', pars=[
            Par(name='r0_struct',value=2.95988, min=1.0, max=20.0,  weight=weight, flag=True),
            Par(name='a_struct', value=2.1391,  min=1,   max=50.0, weight=weight, flag=True),
            Par(name='h_struct', value=7.0, min=-np.inf, max=np.inf, weight=weight, flag=True),
            Par(name='r0_apoly1_struct', value=0.0,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='r0_apoly2_struct', value=0.0,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='r0_apoly3_struct', value=0.0,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='a_apoly1_struct', value=0.0,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='a_apoly2_struct', value=0.0,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='a_apoly3_struct', value=0.0,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='apoly0_struct', value=1.0,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='apoly1_struct', value=0.0,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='apoly2_struct', value=0.0, min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='apoly3_struct', value=0.0, min=-np.inf, weight=weight, max=np.inf, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=0.0,  min=-30.0,   max=30.0,  flag=True),
        ])
        
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        r_gauss = params['r_gauss'].get_value()
        a_gauss = params['a_gauss'].get_value()
        gauss_r1 = np.exp(-a_gauss*(r1_au-r_gauss)**2)
        gauss_r2 = np.exp(-a_gauss*(r2_au-r_gauss)**2)
        gauss = gauss_r1*gauss_r2
                
        # get structural parameters
        r0_struct = params['r0_struct'].get_value()
        a_struct = params['a_struct'].get_value()
        h_struct = params['h_struct'].get_value()
        r0_apoly1_struct = params['r0_apoly1_struct'].get_value()
        r0_apoly2_struct = params['r0_apoly2_struct'].get_value()
        r0_apoly3_struct = params['r0_apoly3_struct'].get_value()
        a_apoly1_struct = params['a_apoly1_struct'].get_value()
        a_apoly2_struct = params['a_apoly2_struct'].get_value()
        a_apoly3_struct = params['a_apoly3_struct'].get_value()
        apoly0_struct = params['apoly0_struct'].get_value()
        apoly1_struct = params['apoly1_struct'].get_value()
        apoly2_struct = params['apoly2_struct'].get_value()
        apoly3_struct = params['apoly3_struct'].get_value()
        r0_struct = r0_struct*(1+r0_apoly1_struct*alpha_part+\
                                 r0_apoly2_struct*alpha_part**2+\
                                 r0_apoly3_struct*alpha_part**3)
        a_struct = a_struct*(1+a_apoly1_struct*alpha_part+\
                               a_apoly2_struct*alpha_part**2+\
                               a_apoly3_struct*alpha_part**3)
        #struct_r1 = 1/2*(np.tanh(a_struct*(r1_au-r0_struct)/r0_struct)+1)
        #struct_r2 = 1/2*(np.tanh(a_struct*(r2_au-r0_struct)/r0_struct)+1)
        struct_r1 = 1/2*(np.tanh(a_struct*(r1_au-r0_struct))+1)
        struct_r2 = 1/2*(np.tanh(a_struct*(r2_au-r0_struct))+1)
        struct = h_struct*struct_r1*struct_r2
        struct *= (apoly0_struct+apoly1_struct*alpha_part+\
                   apoly2_struct*alpha_part**2+apoly3_struct*alpha_part**3)
                           
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # get final result
        res = poly*gauss+struct+global_shift
        
        # set components
        self.__components__['gauss'] = gauss
        self.__components__['poly'] = poly
        self.__components__['polygauss'] = poly*gauss
        self.__components__['struct'] = struct 
        self.__components__['globshift'] = global_shift
            
        return res

class DBOCModel_666_22(Model): # MODEL TO FIT DBOC(2,2) FOR 666 ISOTOPOLOGUE
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__components__ = {}
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
            #[0,     0], # 0   
            
            [1,     0], # 1   
            [0,     1], # 1   
            
            [1,     1], # 2   
            #[0,     2], # 2   
            [2,     0], # 2   
            
            [3,     0], # 3
            #[2,     1], # 3
            #[1,     2], # 3
            #[0,     3], # 3
            
            #[4,     0], # 4
            #[3,     1], # 4
            #[2,     2], # 4
            #[1,     3], # 4
            #[0,     4], # 4
        ])
                
        # polynomial parameters, active by default, no bounds
        weight = 0.0
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,weight=weight,flag=True)
        
        # gaussian parameters, with bounds
        weight = 0.0
        self.__params__.append(group='gaussian', pars=[
            Par(name='center_rgauss', value=1.5, min=-50.0, max=50.0, weight=weight, flag=True),
            Par(name='slope_rgauss', value=0.5, min=0.4, max=20.0, weight=weight, flag=True),
            Par(name='center_agauss', value=107.0, min=-180.0, max=180.0, weight=weight, flag=True),
            Par(name='slope_agauss', value=4.1, min=2.0, max=10.0, weight=weight, flag=True),
            Par(name='mul_gauss', value=35.0, min=0.0, max=100.0, weight=weight, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=15.0,  min=10.0,   max=20.0,  flag=True),
        ])
        
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        center_rgauss = params['center_rgauss'].get_value()
        slope_rgauss = params['slope_rgauss'].get_value()
        center_agauss = params['center_agauss'].get_value()
        slope_agauss = params['slope_agauss'].get_value()
        mul_gauss = params['mul_gauss'].get_value()
        gauss_r1 = np.exp(-slope_rgauss*((r1_au-center_rgauss)/center_rgauss)**2)
        gauss_r2 = np.exp(-slope_rgauss*((r2_au-center_rgauss)/center_rgauss)**2)
        gauss_alpha = np.exp(-slope_agauss*((alpha_deg-center_agauss)/center_agauss)**2)
        gauss = mul_gauss*gauss_r1*gauss_r2*gauss_alpha
                                           
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # structure is set to zero
        struct = 0.0
        
        # get final result
        res = global_shift-(1+poly)*gauss
        
        # set components
        self.__components__['gauss'] = gauss
        self.__components__['poly'] = poly
        self.__components__['polygauss'] = poly*gauss
        self.__components__['struct'] = struct 
        self.__components__['globshift'] = global_shift
            
        return res

class DBOCModel_666_44(Model): # MODEL TO FIT DBOC(4,4) FOR 666 ISOTOPOLOGUE
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__components__ = {}
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
            #[0,     0], # 0   
            
            [1,     0], # 1   
            [0,     1], # 1   
            
            [1,     1], # 2   
            #[0,     2], # 2   
            [2,     0], # 2   
            
            #[3,     0], # 3
            #[2,     1], # 3
            #[1,     2], # 3
            #[0,     3], # 3
            
            #[4,     0], # 4
            #[3,     1], # 4
            #[2,     2], # 4
            #[1,     3], # 4
            #[0,     4], # 4
        ])
                
        # polynomial parameters, active by default, no bounds
        weight = 0.0
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,weight=weight,flag=True)
        
        # gaussian parameters, with bounds
        weight = 0.0
        self.__params__.append(group='gaussian', pars=[
            Par(name='center_rgauss', value=1.5, min=-50.0, max=50.0, weight=weight, flag=True),
            Par(name='slope_rgauss', value=0.5, min=0.4, max=20.0, weight=weight, flag=True),
            Par(name='center_agauss', value=107.0, min=-180.0, max=180.0, weight=weight, flag=True),
            Par(name='slope_agauss', value=4.1, min=2.0, max=10.0, weight=weight, flag=True),
            Par(name='mul_gauss', value=35.0, min=0.0, max=100.0, weight=weight, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=15.0,  min=10.0,   max=20.0,  flag=True),
        ])
        
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        center_rgauss = params['center_rgauss'].get_value()
        slope_rgauss = params['slope_rgauss'].get_value()
        center_agauss = params['center_agauss'].get_value()
        slope_agauss = params['slope_agauss'].get_value()
        mul_gauss = params['mul_gauss'].get_value()
        gauss_r1 = np.exp(-slope_rgauss*((r1_au-center_rgauss)/center_rgauss)**2)
        gauss_r2 = np.exp(-slope_rgauss*((r2_au-center_rgauss)/center_rgauss)**2)
        gauss_alpha = np.exp(-slope_agauss*((alpha_deg-center_agauss)/center_agauss)**2)
        #gauss_alpha = 1
        gauss = mul_gauss*gauss_r1*gauss_r2*gauss_alpha
                                           
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # structure is set to zero
        struct = 0.0
        
        # get final result
        res = global_shift-(1+poly)*gauss
        
        # set components
        self.__components__['gauss'] = gauss
        self.__components__['poly'] = poly
        self.__components__['polygauss'] = poly*gauss
        self.__components__['struct'] = struct 
        self.__components__['globshift'] = global_shift
            
        return res

class DBOCModel_666_66(Model): # LAME MODEL TO FIT DBOC(6,6) FOR 666 ISOTOPOLOGUE
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__components__ = {}
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
            #[0,     0], # 0   
            
            [1,     0], # 1   
            [0,     1], # 1   
            
            [1,     1], # 2   
            [0,     2], # 2   
            [2,     0], # 2   
            
            #[3,     0], # 3
            #[2,     1], # 3
            #[1,     2], # 3
            #[0,     3], # 3
            
            #[4,     0], # 4
            #[3,     1], # 4
            #[2,     2], # 4
            #[1,     3], # 4
            #[0,     4], # 4
        ])
                
        # polynomial parameters, active by default, no bounds
        weight = 0.0
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,weight=weight,flag=True)
        
        # gaussian parameters, with bounds
        weight = 0.0
        self.__params__.append(group='gaussian', pars=[
            Par(name='center_rgauss', value=2.5, min=2.0, max=3.0, weight=weight, flag=True),
            Par(name='slope_rgauss', value=10.0, min=5.0, max=20.0, weight=weight, flag=True),
            Par(name='center_agauss', value=107.0, min=-180.0, max=180.0, weight=weight, flag=True),
            Par(name='slope_agauss', value=4.1, min=2.0, max=10.0, weight=weight, flag=True),
            Par(name='mul_gauss', value=20.0, min=0.0, max=100.0, weight=weight, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=10.0,  min=5.0,   max=20.0,  flag=True),
        ])
        
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        center_rgauss = params['center_rgauss'].get_value()
        slope_rgauss = params['slope_rgauss'].get_value()
        center_agauss = params['center_agauss'].get_value()
        slope_agauss = params['slope_agauss'].get_value()
        mul_gauss = params['mul_gauss'].get_value()
        gauss_r1 = np.exp(-slope_rgauss*((r1_au-center_rgauss)/center_rgauss)**2)
        gauss_r2 = np.exp(-slope_rgauss*((r2_au-center_rgauss)/center_rgauss)**2)
        gauss_alpha = np.exp(-slope_agauss*((alpha_deg-center_agauss)/center_agauss)**2)
        #gauss_alpha = 1
        gauss = mul_gauss*gauss_r1*gauss_r2*gauss_alpha
                                           
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # structure is set to zero
        struct = 0.0
        
        # get final result
        res = global_shift-(1+poly)*gauss
        
        # set components
        self.__components__['gauss'] = gauss
        self.__components__['poly'] = poly
        self.__components__['polygauss'] = poly*gauss
        self.__components__['struct'] = struct 
        self.__components__['globshift'] = global_shift
            
        return res

class DBOCModel_666_66_2(Model): # model to refit the full DBOC correction for 66; direct copy of DBOCModel_666_66 with some initial parameter changes
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__components__ = {}
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
            #[0,     0], # 0   
            
            [1,     0], # 1   
            [0,     1], # 1   
            
            [1,     1], # 2   
            [0,     2], # 2   
            [2,     0], # 2   
            
            #[3,     0], # 3
            #[2,     1], # 3
            #[1,     2], # 3
            #[0,     3], # 3
            
            #[4,     0], # 4
            #[3,     1], # 4
            #[2,     2], # 4
            #[1,     3], # 4
            #[0,     4], # 4
        ])
                
        # polynomial parameters, active by default, no bounds
        weight = 0.0
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,weight=weight,flag=True)
        
        # gaussian parameters, with bounds
        weight = 0.0
        self.__params__.append(group='gaussian', pars=[
            Par(name='center_rgauss', value=2.4, min=2.0, max=3.0, weight=weight, flag=True),
            Par(name='slope_rgauss', value=10.0, min=5.0, max=20.0, weight=weight, flag=True),
            Par(name='center_agauss', value=117.0, min=-180.0, max=180.0, weight=weight, flag=True),
            Par(name='slope_agauss', value=4.1, min=2.0, max=10.0, weight=weight, flag=True),
            Par(name='mul_gauss', value=15.0, min=0.0, max=100.0, weight=weight, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=15.0,  min=5.0,   max=20.0,  flag=True),
        ])
        
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        center_rgauss = params['center_rgauss'].get_value()
        slope_rgauss = params['slope_rgauss'].get_value()
        center_agauss = params['center_agauss'].get_value()
        slope_agauss = params['slope_agauss'].get_value()
        mul_gauss = params['mul_gauss'].get_value()
        gauss_r1 = np.exp(-slope_rgauss*((r1_au-center_rgauss)/center_rgauss)**2)
        gauss_r2 = np.exp(-slope_rgauss*((r2_au-center_rgauss)/center_rgauss)**2)
        gauss_alpha = np.exp(-slope_agauss*((alpha_deg-center_agauss)/center_agauss)**2)
        #gauss_alpha = 1
        gauss = mul_gauss*gauss_r1*gauss_r2*gauss_alpha
                                           
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # structure is set to zero
        struct = 0.0
        
        # get final result
        res = global_shift-(1+poly)*gauss
        
        # set components
        self.__components__['gauss'] = gauss
        self.__components__['poly'] = poly
        self.__components__['polygauss'] = poly*gauss
        self.__components__['struct'] = struct 
        self.__components__['globshift'] = global_shift
            
        return res
        
class DBOCModel_test2(Model): # SANDBOX MODEL FOR TESTING PURPOSES
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__components__ = {}
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
            #[0,     0], # 0   
            
            [1,     0], # 1   
            [0,     1], # 1   
            
            [1,     1], # 2   
            #[0,     2], # 2   
            [2,     0], # 2   
            
            #[3,     0], # 3
            #[2,     1], # 3
            #[1,     2], # 3
            #[0,     3], # 3
            
            #[4,     0], # 4
            #[3,     1], # 4
            #[2,     2], # 4
            #[1,     3], # 4
            #[0,     4], # 4
        ])
                
        # polynomial parameters, active by default, no bounds
        weight = 0.0
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,weight=weight,flag=True)
        
        # gaussian parameters, with bounds
        weight = 0.0
        self.__params__.append(group='gaussian', pars=[
            Par(name='center_rgauss', value=1.5, min=-50.0, max=50.0, weight=weight, flag=True),
            Par(name='slope_rgauss', value=0.5, min=0.4, max=20.0, weight=weight, flag=True),
            Par(name='center_agauss', value=107.0, min=-180.0, max=180.0, weight=weight, flag=True),
            Par(name='slope_agauss', value=4.1, min=2.0, max=10.0, weight=weight, flag=True),
            Par(name='mul_gauss', value=35.0, min=0.0, max=100.0, weight=weight, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=15.0,  min=10.0,   max=20.0,  flag=True),
        ])
        
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        center_rgauss = params['center_rgauss'].get_value()
        slope_rgauss = params['slope_rgauss'].get_value()
        center_agauss = params['center_agauss'].get_value()
        slope_agauss = params['slope_agauss'].get_value()
        mul_gauss = params['mul_gauss'].get_value()
        gauss_r1 = np.exp(-slope_rgauss*((r1_au-center_rgauss)/center_rgauss)**2)
        gauss_r2 = np.exp(-slope_rgauss*((r2_au-center_rgauss)/center_rgauss)**2)
        gauss_alpha = np.exp(-slope_agauss*((alpha_deg-center_agauss)/center_agauss)**2)
        #gauss_alpha = 1
        gauss = mul_gauss*gauss_r1*gauss_r2*gauss_alpha
                                           
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # structure is set to zero
        struct = 0.0
        
        # get final result
        res = global_shift-(1+poly)*gauss
        
        # set components
        self.__components__['gauss'] = gauss
        self.__components__['poly'] = poly
        self.__components__['polygauss'] = poly*gauss
        self.__components__['struct'] = struct 
        self.__components__['globshift'] = global_shift
            
        return res
        
class DBOCModel_3(Model):
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
        #   r_symm  alpha
            [0,     0], # 0
            
            [1,     0], # 1
            [0,     1], # 1
            
            [1,     1], # 2
            [0,     2], # 2
            [2,     0], # 2
        ])
                
        # polynomial parameters, active by default, no bounds
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,flag=True)
        
        # gaussian parameters, with bounds
        self.__params__.append(group='gaussian', pars=[
            Par(name='r_gauss', value=2.01, min=2.0, max=4.0,  flag=True),
            Par(name='a_gauss', value=9.99964, min=2.0, max=10.0, flag=True),
        ])
        
        # structure parameters, with bounds
        weight = 10.0
        self.__params__.append(group='structure', pars=[
            Par(name='r0_struct',value=2.95988, min=1.0, max=5.0,  weight=weight, flag=True),
            Par(name='a_struct', value=2.1391,  min=1,   max=50.0, weight=weight, flag=True),
            Par(name='h_struct', value=7.0, min=-np.inf, max=np.inf, weight=weight, flag=False),
            Par(name='apoly0_struct', value=3.01235,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='apoly1_struct', value=1.06193,  min=-np.inf, weight=weight, max=np.inf, flag=True),
            Par(name='apoly2_struct', value=0.269482, min=-np.inf, weight=weight, max=np.inf, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=0.0,  min=-30.0,   max=30.0,  flag=False),
        ])
        
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        r_gauss = params['r_gauss'].get_value()
        a_gauss = params['a_gauss'].get_value()
        gauss_r1 = np.exp(-a_gauss*(r1_au-re_au)**2)
        gauss_r2 = np.exp(-a_gauss*(r2_au-re_au)**2)
        gauss = gauss_r1*gauss_r2
        
        # get structural parameters
        r0_struct = params['r0_struct'].get_value()
        a_struct = params['a_struct'].get_value()
        h_struct = params['h_struct'].get_value()
        apoly0_struct = params['apoly0_struct'].get_value()
        apoly1_struct = params['apoly1_struct'].get_value()
        apoly2_struct = params['apoly2_struct'].get_value()
        struct_r1 = 1/2*(np.tanh(a_struct*(r1_au-r0_struct)/r0_struct)+1)
        struct_r2 = 1/2*(np.tanh(a_struct*(r2_au-r0_struct)/r0_struct)+1)
        struct = h_struct*struct_r1*struct_r2
        struct *= (apoly0_struct+apoly1_struct*alpha_part+apoly2_struct*alpha_part**2)
        
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # get final result
        res = poly*gauss+struct+global_shift
        
        return res


        
class TestModelFortran(modfort.ModelFortran):
    """
    Sample Fortran model.
    """
    def __init__(self):
        self.__name__ = 'my_test_1'
        self.__maxpower__ = 10
        self.__powers__ = np.array(generate_powers(self.__maxpower__))
        self.__params__ = Parameters(N=len(self.__powers__))
        self.__data__ = modfort.Data(
            modfort.FortranValue('Shift','double precision',100.0),
            modfort.FortranMatrix('Powers','integer',self.__powers__),
            )
        self.__code__ = """
        subroutine foo(res,p,x,y)
        integer :: i
        !res = p(1)*x+p(2)*y+p(3)*x**2+p(4)*y**2+Shift
        res = 0.0d0
        do i=1,np__
            res = res + p(i)*x**Powers(i,1)*y**Powers(i,2)
        end do
        !print *,'p',p
        !print *,'x',x,'y',y,'res',res
        end subroutine
        """
        super().__init__()
            
class PolynomialModelFortran():
    pass
    

# ========================================        
# ========================================        
# MODELS FOR FITTING THE 888 ISOTOPOLOGUE        
# ========================================        
# ========================================        

class DBOCModel_888_66_2(Model): # model to refit the full DBOC correction for 66; direct copy of DBOCModel_666_66_2
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__components__ = {}
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
            #[0,     0], # 0   
            
            [1,     0], # 1   
            [0,     1], # 1   
            
            [1,     1], # 2   
            [0,     2], # 2   
            [2,     0], # 2   
            
            #[3,     0], # 3
            #[2,     1], # 3
            #[1,     2], # 3
            #[0,     3], # 3
            
            #[4,     0], # 4
            #[3,     1], # 4
            #[2,     2], # 4
            #[1,     3], # 4
            #[0,     4], # 4
        ])
                
        # polynomial parameters, active by default, no bounds
        weight = 0.0
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,weight=weight,flag=True)
        
        # gaussian parameters, with bounds
        weight = 0.0
        self.__params__.append(group='gaussian', pars=[
            Par(name='center_rgauss', value=2.4, min=2.0, max=3.0, weight=weight, flag=True),
            Par(name='slope_rgauss', value=10.0, min=5.0, max=20.0, weight=weight, flag=True),
            Par(name='center_agauss', value=117.0, min=-180.0, max=180.0, weight=weight, flag=True),
            Par(name='slope_agauss', value=4.1, min=2.0, max=10.0, weight=weight, flag=True),
            Par(name='mul_gauss', value=15.0, min=0.0, max=100.0, weight=weight, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=15.0,  min=5.0,   max=20.0,  flag=True),
        ])
        
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        center_rgauss = params['center_rgauss'].get_value()
        slope_rgauss = params['slope_rgauss'].get_value()
        center_agauss = params['center_agauss'].get_value()
        slope_agauss = params['slope_agauss'].get_value()
        mul_gauss = params['mul_gauss'].get_value()
        gauss_r1 = np.exp(-slope_rgauss*((r1_au-center_rgauss)/center_rgauss)**2)
        gauss_r2 = np.exp(-slope_rgauss*((r2_au-center_rgauss)/center_rgauss)**2)
        gauss_alpha = np.exp(-slope_agauss*((alpha_deg-center_agauss)/center_agauss)**2)
        #gauss_alpha = 1
        gauss = mul_gauss*gauss_r1*gauss_r2*gauss_alpha
                                           
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # structure is set to zero
        struct = 0.0
        
        # get final result
        res = global_shift-(1+poly)*gauss
        
        # set components
        self.__components__['gauss'] = gauss
        self.__components__['poly'] = poly
        self.__components__['polygauss'] = (1+poly)*gauss
        self.__components__['struct'] = struct 
        self.__components__['globshift'] = global_shift
            
        return res
        
# ========================================        
# ========================================        
# MODELS FOR FITTING THE 777 ISOTOPOLOGUE        
# ========================================        
# ========================================        
        
class DBOCModel_777_66_2(Model): # model to refit the full DBOC correction for 66, iso 777; direct copy of DBOCModel_666_66_2
    """
    Model for DBOC points fitting
    """
    def __init__(self,fast=False):
        self.__components__ = {}
        self.__fast__ = fast
        self.__powers__ = np.array([
        #   r_symm  alpha
            #[0,     0], # 0   
            
            [1,     0], # 1   
            [0,     1], # 1   
            
            [1,     1], # 2   
            [0,     2], # 2   
            [2,     0], # 2   
            
            #[3,     0], # 3
            #[2,     1], # 3
            #[1,     2], # 3
            #[0,     3], # 3
            
            #[4,     0], # 4
            #[3,     1], # 4
            #[2,     2], # 4
            #[1,     3], # 4
            #[0,     4], # 4
        ])
                
        # polynomial parameters, active by default, no bounds
        weight = 0.0
        self.__params__ = Parameters(group='polynom',N=len(self.__powers__),
            prefix='lin_',names=['%d_%d'%tuple(_) for _ in self.__powers__],
            value=0.0,weight=weight,flag=True)
        
        # gaussian parameters, with bounds
        weight = 0.0
        self.__params__.append(group='gaussian', pars=[
            Par(name='center_rgauss', value=2.4, min=2.0, max=3.0, weight=weight, flag=True),
            Par(name='slope_rgauss', value=10.0, min=5.0, max=20.0, weight=weight, flag=True),
            Par(name='center_agauss', value=117.0, min=-180.0, max=180.0, weight=weight, flag=True),
            Par(name='slope_agauss', value=4.1, min=2.0, max=10.0, weight=weight, flag=True),
            Par(name='mul_gauss', value=15.0, min=0.0, max=100.0, weight=weight, flag=True),
        ])

        # global shift parameters, with bounds
        self.__params__.append(group='globshift', pars=[
            Par(name='global_shift', value=15.0,  min=5.0,   max=20.0,  flag=True),
        ])
        
    def __func__(self,params,r1_au,r2_au,alpha_deg):
        # common variables
        re_au = 2.4
        alpha_rad = alpha_deg*np.pi/180.
        alphae_rad = 117.0*np.pi/180.
        r_part = (r1_au+r2_au)/re_au
        alpha_part = (np.cos(alpha_rad)-\
                np.cos(alphae_rad))/np.cos(alphae_rad)
                                
        # get polynomial part
        p = params.get_values(group='polynom')
        if self.__fast__:
            poly = fast_poly2d(p,r_part,alpha_part,self.__powers__)
        else:
            poly = poly2d(p,r_part,alpha_part,self.__powers__)
            
        # get gaussian parameters
        center_rgauss = params['center_rgauss'].get_value()
        slope_rgauss = params['slope_rgauss'].get_value()
        center_agauss = params['center_agauss'].get_value()
        slope_agauss = params['slope_agauss'].get_value()
        mul_gauss = params['mul_gauss'].get_value()
        gauss_r1 = np.exp(-slope_rgauss*((r1_au-center_rgauss)/center_rgauss)**2)
        gauss_r2 = np.exp(-slope_rgauss*((r2_au-center_rgauss)/center_rgauss)**2)
        gauss_alpha = np.exp(-slope_agauss*((alpha_deg-center_agauss)/center_agauss)**2)
        #gauss_alpha = 1
        gauss = mul_gauss*gauss_r1*gauss_r2*gauss_alpha
                                           
        # get global shift
        global_shift = params['global_shift'].get_value()
        
        # structure is set to zero
        struct = 0.0
        
        # get final result
        res = global_shift-(1+poly)*gauss
        
        # set components
        self.__components__['gauss'] = gauss
        self.__components__['poly'] = poly
        self.__components__['polygauss'] = (1+poly)*gauss
        self.__components__['struct'] = struct 
        self.__components__['globshift'] = global_shift
            
        return res



from collections import OrderedDict

from ..base import Config
from . import general, data, model, fit, stat, codegen, plotting, calc

template_modules_list = general, data, model, fit, stat, codegen, plotting, calc

template_modules_dict = OrderedDict()
for mod in template_modules_list:
    template_modules_dict[mod.__name__.split('.')[-1]] = mod

config = Config(*[getattr(module,'Default') for module in template_modules_list])



from ..base import ConfigSection

class Default(ConfigSection):
    __name__ = 'FIT'
    __header__ = 'FITTER OPTIONS'
    __template__ = \
"""
# Weighted / unweighted fit mode.
{weighted_fit}

# Use "rubber" regularization.
{rubber_on}

# Fitting method (trf, lm, basinhopping).
# Available "local" methods:
#    "trf" -> Trust Region Reflective (least squares, bounded).
#    "lm"  -> Levenberg Marquardt (least squares, unbounded).
#    'Nelder-Mead'  -> Gradient-free descent Nelder-Mead method.
#    'Powell'  -> modified Powell algorithm.
#    'CG'  -> conjugate gradient algorithm.
#    'BFGS'  -> BFGS algorithm.
#    'Newton-CG'  -> Newton-CG algorithm.
#    'L-BFGS-B'  -> L-BFGS-B algorithm.
#    'TNC'  -> truncated Newton (TNC) algorithm.
#    'COBYLA'  -> Constrained Optimization BY Linear Approximation.
#    'SLSQP'  -> Sequential Least Squares Programming.
#    'trust-constr'  -> minimize a scalar function subject to constraints.
#    'dogleg'  -> dog-leg trust-region algorithm.
#    'trust-ncg'  -> Newton conjugate gradient trust-region algorithm.
#    'trust-exact'  -> "nearly exact trust-region algorithm"
#    'trust-krylov'  -> "early exact trust-region algorithm"
# Available "global" methods:
#    "basinhopping"  -> local descents from a random starting point (bounded)
#    "anneal" -> Simulated Annealing method
{fitting_method}

# Calculate analytic jacobian (only if ModelSympy is used).
{analytic_jacobian}

# Fit options. Must be given as a list separated by semicolon.
# The valid options for most of the supported fitting methods
# can be found in the corresponding scipy.optimize documentation.  
{fit_options}
"""
    weighted_fit = True
    rubber_on = True
    fitting_method = 'trf'
    analytic_jacobian = True
    fit_options = 'max_nfev=200;'



import numpy as np
from tabulate import tabulate as tab

class Parameters:
    """
    Aggregator for fitting parameters.
    This is not necessary to use in models: one can just use the arrays for parameter.
    """
    def __init__(self,N=0,prefix=None,names=None,
                 values=None,flags=True,mins=None,maxs=None,
                 weights=None,group='default'):
        self.__objects__ = []
        self.__index__ = {}
        self.generate(N=N,prefix=prefix,group=group,
                      names=names,values=values,flags=flags,
                      mins=mins,maxs=maxs,weights=weights)
                      
    def dump(self): # (DE)SERIALIZATION FOR PAPER
        parlist = []
        for par in self.__objects__:
            parlist.append(par.dump())
        return parlist
        
    def load(self,parlist): # (DE)SERIALIZATION FOR PAPER
        self.__objects__ = []
        self.__index__ = {}
        for dct in parlist:
            par = Par(name=dct['name']); par.load(dct)
            self.append(par)        
        
    def getcode(self,indent=4,attrs=None): # (DE)SERIALIZATION FOR PAPER
        buf = ''
        for par in self.__objects__:
            buf += (' '*indent+'%s,\n')%par.getcode(attrs=attrs)
        return '[\n%s]'%buf
        
    def __iter__(self):
        return iter(self.__objects__) # strange peculiarity of Python 3
        
    def __len__(self):
        return len(self.__objects__)
                      
    @property
    def __npar__(self):
        return len(self.__objects__)
        
    def __getitem__(self,name):
        """
        Get parameter object by its name.
        """
        return self.__objects__[self.__index__[name]]
        
    def append(self,pars,group='default'):
        """
        Append a parameter to the set. Pars can be either a parameter,
        or a list of parameters.
        """
        if type(pars) not in {list,tuple,Parameters}:
            pars = [pars]
        for par in pars:
            par.__group__ = group
            if par.__name__ not in self.__index__:
                self.__index__[par.__name__] = len(self.__objects__)
                self.__objects__.append(par)
            else:
                self.__objects__[self.__index__[par.__name__]] = par
                        
    def generate(self,N=None,prefix=None,names=None,group='default',
                 values=None,flags=True,mins=None,maxs=None,weights=None):
        """
        Generate n parameters with predefined values.
        """
        if N is None and names is None:
            raise Exception('both N and names are None')
        if prefix is None: prefix = ''
        if names is None:
            # generate names:
            names = [prefix+'%d'%i for i in range(N)]
        else:
            if len(names)!=N:
                raise Exception('len(names)!=N')
        N = len(names)
        if type(weights) not in {list,tuple}:
            weights = [weights for i in range(N)]
        if type(values) not in {list,tuple}:
            values = [values for i in range(N)]
        if type(flags) not in {list,tuple}:
            flags = [flags for i in range(N)]
        if type(mins) not in {list,tuple}:
            mins = [mins for i in range(N)]
        if type(maxs) not in {list,tuple}:
            maxs = [maxs for i in range(N)]
        for name,weight,value,flag,min,max in \
                zip(names,weights,values,flags,mins,maxs):
            self.append(Par(name=name,value=value,flag=flag,
                    min=min,max=max,weight=weight),
                    group=group)

    def get_groups(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array([self.__objects__[i].__group__ for i in ii])
                    
    def get_names(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array([self.__objects__[i].__name__ for i in ii])
        
    def get_values(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array([self.__objects__[i].__value__ for i in ii])

    def set_values(self,values,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)        
        for i,val in zip(ii,values):
            self.__objects__[i].__value__ = val
        
    def get_bounds(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return [self.__objects__[i].get_bounds() for i in ii]
        
    def get_weights(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array([self.__objects__[i].__weight__ for i in ii])
    
    def get_flags(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array([self.__objects__[i].__flag__ for i in ii])
        
    def get_nonzero_weights_and_values(self):
        # THIS CAN HELP FITTING BY REMOVING THE ZERO-WEIGHTED COMPONENTS IN THE RUBBER
        # TO BE IMPLEMENTED
        raise NotImplementedError
        
    def get_index(self,active_only=False,group=None):
        # REDO!!!
        if active_only and group:
            ii = [i for i,p in enumerate(self.__objects__) if p.__group__==group and p.__flag__]
        elif active_only and not group: 
            ii = [i for i,p in enumerate(self.__objects__) if p.__flag__]  
        elif not active_only and group:
            ii = [i for i,p in enumerate(self.__objects__) if p.__group__==group]            
        elif not active_only and not group:
            ii = [i for i,p in enumerate(self.__objects__)]
        return ii
        
    def __repr__(self):
        groups = self.get_groups()
        names = self.get_names()
        values = self.get_values()
        bounds = self.get_bounds()
        weights = self.get_weights()
        flags = self.get_flags()
        indices = []; i = 0
        for flag in flags:
            if flag:
                indices.append(i); i += 1
            else:
                indices.append('')
        return tab(zip(indices,groups,names,values,bounds,weights,flags),
                headers=('#','group','names','values','bounds','weights','flags'))

def dumpval(val):
    """ Simple alternative to json.dumps for VERY limited set of value types:
        1) only scalar types are supported: int, float, str, bool
    """
    if type(val) is str:
        return '"%s"'%val
    else:
        return str(val)
                
class Par:
    """
    Fitting parameter storing information about initial values, bounds, rubber weight, and active flag.
    Each parameter has its unique name.
    """
    MAP = { # (DE)SERIALIZATION FOR PAPER
        'name': '__name__',
        'group': '__group__',
        'value': '__value__',
        'flag': '__flag__',
        'min': '__min__',
        'max': '__max__',
        'weight': '__weight__',
    }
    
    def __init__(self,name,value=None,flag=None,min=None,max=None,weight=None,group=None):
        self.__name__ = name
        self.__group__ = 'default' if group is None else group
        self.__value__ = np.random.random() if value is None else value
        self.__flag__ = False if flag is None else flag
        self.__min__ = -np.inf if min is None else min
        self.__max__ = np.inf if max is None else max
        self.__weight__ = 0.0 if weight is None else weight
        
    def dump(self): # (DE)SERIALIZATION FOR PAPER
        keys = set(self.MAP)
        return {name:getattr(self,self.MAP[name]) for name in keys}
        
    def load(self,dct): # (DE)SERIALIZATION FOR PAPER
        for name in dct:
            if name in self.MAP: setattr(self,self.MAP[name],dct[name])
                
    def getcode(self,attrs=None): # LIMITED AND DISPOSABLE ALTERNATIVE OF THE PROPER DUMP FUNCTIONAL; SHOULD BE USED FOR THE OZONE PAPER ONLY!!!
        if attrs is None: attrs = list(self.MAP.keys())
        dct = {name:getattr(self,self.MAP[name]) for name in attrs}
        return '{%s}'%(', '.join(['"%s":%s'%(name,dumpval(dct[name])) for name in dct]))
                              
    def get_value(self):
        return self.__value__
        
    def set_value(self,value):
        self.__value__ = value
        
    def get_flag(self):
        return self.__flag__
        
    def set_flag(self,flag):
        self.__flag__ = flag        
        
    def get_bounds(self):
        return [self.__min__,self.__max__]
        
    def set_bounds(self,bnds):
        self.__min__ = bnds[0]
        self.__max__ = bnds[1]
        
    def get_min(self):
        return self.__min__
        
    def set_min(self,min):
        self.__min__ = min

    def get_max(self):
        return self.__max__
        
    def set_max(self,max):
        self.__max__ = max
        
    def get_weight(self):
        return self.__weight__
        
    def set_weight(self,wht):
        self.__weight__ = wht
        
    def get_name(self):
        return self.__name__    
        
    def __repr__(self):
        return '{}:{}:{}({}/[{}:{}])*{}'.format(self.__group__,
        self.__name__,self.__flag__,self.__value__,self.__min__,
        self.__max__,self.__weight__)



import re
import copy
import inspect
from collections import OrderedDict
import configparser

def __get_members_recursive__(cls,members):
    # get stuff from current class
    for attribute in cls.__dict__.keys():
        value = getattr(cls, attribute)
        if attribute[:2] != '__' and not callable(value):
            if attribute not in members:
                members[attribute] = value
    # get stuff from base class recursively
    bases = cls.__bases__
    if len(bases)>1:
        raise Exception(
            'ConfigSection instance cannot have more than one parents')
    base = bases[0]
    if base is not ConfigSection:
        __get_members_recursive__(base,members)
        
class ConfigSection:
    """ 
    Class for the config part. 
    Essential class parameters:
        __name__: config section name 
        __header__: short header telling what the section izs about
        __template__: the config representation of the section
    The parameters of the section are given within template
    """
    
    def __init__(self,**kwargs):
        # set members 
        self.__set_members__(kwargs=kwargs)
        # set defaults (absent members)
        template_keys = self.__class__.__get_keys_from_template__()
        members = self.__get_members__()
        absent_keys = set(template_keys)-set(members.keys())
        for key in absent_keys:
            setattr(self,key,None)
    
    @classmethod
    def __get_keys_from_template__(cls):
        return re.findall('{([a-zA-Z][a-zA-Z0-9_]*)}',cls.__template__)
        
    @classmethod
    def __get_members_from_class__(cls):
        members = {}
        __get_members_recursive__(cls,members)
        return members

    def __get_members_from_instance__(self):
        members = {}
        for attribute in self.__dict__.keys():
            value = getattr(self, attribute)
            if attribute[:2] != '__' and not callable(value):
                members[attribute] = value
        return members

    def __get_members__(self):
        members = copy.deepcopy(self.__get_members_from_class__())
        members.update(copy.deepcopy(self.__get_members_from_instance__()))
        return members

    def __set_members__(self,kwargs,ignore_empty_values=True):
        allowed_members = self.__get_members_from_instance__()
        allowed_members = set(self.__get_keys_from_template__()).\
            union(allowed_members.keys())
        for kwarg in kwargs:
            if kwarg in allowed_members:
                if ignore_empty_values and not kwargs[kwarg]:
                    continue
                setattr(self,kwarg,kwargs[kwarg])
            else:
                raise Exception('the key "%s" is not in dict'%kwarg)

    @property
    def buffer(self):
        template_keys = self.__class__.__get_keys_from_template__()
        members = self.__get_members__()
        # add keys existing as members
        dct = {key:'{key}: {value}'.\
            format(key=key,value=members[key] if members[key] is not None else '') \
                for key in members if key in template_keys}
        # make a header part
        head = '{binding}\n# {header} #\n{binding}\n[{name}]\n'.\
            format(binding='#'*(len(self.__header__)+4),
                header=self.__header__,name=self.__name__)
        # all done
        return head + self.__class__.__template__.format(**dct) + '\n'
        
    def get(self,key,default=None):
        try:
            return getattr(self,key)
        except AttributeError:
            return default
        
    def print(cls):
        print(cls.buffer)
        
    def __getitem__(self,key):
        return getattr(self,key)

    def __setitem__(self,key,value):
        allowed_members = self.__get_members__()
        if key in allowed_members:
            setattr(self,key,value)
        else:            
            raise Exception('the key "%s" is not in dict'%key)

class Config:
    """ Class for the config. """
    
    def __init__(self,*sections):
        self.__sections__ = OrderedDict()
        for section in sections:
            if type(section) is type:
                if issubclass(section,ConfigSection):
                    section = section()
                else:
                    raise Exception(
                        'wrong class supplied: %s'%str(section))
            self.__sections__[section.__name__] = section
    
    def setpars(self,pars,ignore_empty_values=True):
        for section_name in pars:
            pp = pars[section_name]
            self.__sections__[section_name].__set_members__(
                kwargs=pp,ignore_empty_values=ignore_empty_values)
            
    def merge_section(self,section,ignore_empty_values=True):
        if type(section) is type:
            section = section()
        dct = {section.__name__:section.__get_members__()}
        self.setpars(dct,ignore_empty_values=ignore_empty_values)
        
    def merge_config(self,config,ignore_empty_values=True):
        for section_name in config.__sections__:
            section = config.__sections__[section_name]
            if section_name not in self.__sections__:
                self.__sections__[section_name] = section.__class__()
            self.merge_section(section,ignore_empty_values=ignore_empty_values)
    
    @property
    def buffer(self):
        buf = ''
        for name in self.__sections__:
            section = self.__sections__[name]
            buf += section.buffer
        return buf
        
    def print(self):
        print(self.buffer)

    def __getitem__(self,section_name):
        return self.__sections__[section_name]
        
    def save(self,filename):
        with open(filename,'w') as f:
            f.write(self.buffer)
        
    def load(self,filename,ignore_empty_values=False):
        conf_ = configparser.ConfigParser(allow_no_value=True)
        conf_.read(filename)
        sections = conf_.sections()
        for section in sections:
            pardict = dict(conf_.items(section))
            self[section].__set_members__(kwargs=pardict,
                ignore_empty_values=ignore_empty_values)
 
def print_help(modules,more=False):
    """ 
    Explore modules with the config section templates, 
    and print their docstrings 
    """
    print(r'||||||||||||||||||||||||||||||||')
    print(r'||   AVAILABLE TEMPLATES:     ||')
    print(r'||||||||||||||||||||||||||||||||')

    for module_name in modules:
                
        module = modules[module_name]
                
        if more: print('\n==============================================')
        print(module.__name__)
        if more: print('==============================================\n')
        
        for obj_name in dir(module):
            obj = getattr(module,obj_name)
            if type(obj) is type and \
                obj is not ConfigSection and \
                issubclass(obj,ConfigSection):
                
                if more: print('///////////////////////////////////')
                print('   ',obj.__name__)
                if more: print(obj.__doc__)
                if more: print()


import os
import sys
import copy
import subprocess

import numpy as np
import sympy as sy
#import symengine as sy # should be faster than sympy, but not entirely compatible
import numba as nb

from tabulate import tabulate as tab

from .base import Model

def subprocess_run_crossplatform(command,*args,**kwargs):
    """ Wrapper for subprocess.run to enable basic cross-platform execution.
    ==============
    sys.platform:
    ==============
    Linux 	'linux'
    Windows 	'win32'
    Windows/Cygwin 	'cygwin'
    Mac OS X 	'darwin'
    """
    if type(command) not in (list,tuple):
        command = [command]
    exefile = command[0]
    if sys.platform in ['linux','darwin','cygwin']:
        exefile = './'+exefile
    subprocess.run([exefile]+command[1:],*args,**kwargs)
        
# ===============================================================================
# ================================ MODEL SYMPY ==================================
# ===============================================================================

# HOW TO GET THE VAR NAMES OF THE FUNCTION:
# https://stackoverflow.com/questions/582056/getting-list-of-parameter-names-inside-python-function

def get_argnames(foo):
    argcount = foo.__code__.co_argcount
    argnames = foo.__code__.co_varnames[:argcount]
    return argnames

def gradient(scalar_function,variables):
    """https://stackoverflow.com/questions/21166771/is-there-a-vectorized-way-to-calculate-the-gradient-in-sympy?rq=1"""
    matrix_scalar_function = sy.Matrix([scalar_function])
    return matrix_scalar_function.jacobian(sy.Matrix([variables]))    

def create_pots_interface():
    """ 
        Create POTS interface for the DVR program only.
        POTS is calling PES. Generally, this should not depend on the model
        and custom parameter usage, such as nuclear masses.
    """
    
    pots_source = \
    'subroutine pots(vout,r1,r2,alpha)' + '\n' + \
    '' + '\n' + \
    'implicit none' + '\n' + \
    '' + '\n' + \
    '! ====================================================================================' + '\n' + \
    '! POTS is the interface between the DVR3D code and the user defined PES routine PES' + '\n' + \
    '!' + '\n' + \
    '! All data are transferred into POTS either through the argument list (r1,r2,alpha) or' + '\n' + \
    '! common blocks' + '\n' + \
    '!' + '\n' + \
    '! ====================================================================================' + '\n' + \
    '' + '\n' + \
    '! This particular case is the interface with the PS mass dependent water PES ' + '\n' + \
    '' + '\n' + \
    'double precision, intent(in) :: r1    ! first bondlength [BOHR]' + '\n' + \
    'double precision, intent(in) :: r2    ! second bondlength [BOHR]' + '\n' + \
    'double precision, intent(in) :: alpha ! bondangle [RADIAN]' + '\n' + \
    'double precision, intent(out) :: vout    ! returned PES value [AU]' + '\n' + \
    '' + '\n' + \
    '! ==================================================================================' + '\n' + \
    '! ==================================================================================' + '\n' + \
    '' + '\n' + \
    'common /mass/ xmass(3) ! vib. masses (must be in in a.u.)' + '\n' + \
    'real*8 xmass' + '\n' + \
    '' + '\n' + \
    'double precision pes' + '\n' + \
    '' + '\n' + \
    '! call PES function to calculate value of PES in' + '\n' + \
    '! the geometry point {r1,r2,alpha} ' + '\n' + \
    '' + '\n' + \
    'vout = pes(.TRUE.,0,r1,r2,alpha,xmass)' + '\n' + \
    '' + '\n' + \
    'end'
    
    with open('pots.f90','w') as fil:
        fil.write(pots_source)

def create_pes_interface(model):
    """ 
        Create an interface for the VTET/DVR programs     
        Source file must correspond to the models units.
        
        Note that, depending on the "au" parameters,
        the units of the input arguments (and output value),
        supplied in the "pes" interface, are changing:
            if au=true: r supplied in bohr, pes expected in hatree
            if au=false: r supplied in angstroem, pes expected in wavenumbers
        
        In both cases alpha supplied in radians.
    """
    units = model.__units__()
    input_units = units['input']
    output_units = units['output']
    input_names = model.__input_names__
    
    def convert_units_pes_to_model(quantity,model_unit):
        """
        All parameters are strings.
        quantity - paramter to convert units for
        model_units - units of the input in Python model __func__
        """
        if model_unit.lower() in ['bohr','au','a.u.']:
            return quantity
        elif model_unit.lower() in ['ang','ang.','a','angstroem','angstrem']:
            return quantity+'*BOHR_TO_ANGSTROEM'
        elif model_unit.lower() in ['rad','rad.','radian','radians']:
            return quantity
        elif model_unit.lower() in ['deg','deg.','degree','degrees']:
            return quantity+'/DEGREE_TO_RADIAN'
        elif model_unit.lower() in ['hartree']:
            return quantity+'/WAVENUMBER_TO_HARTREE'
        elif model_unit.lower() in ['cm-1','wavenumber','wavenumbers']:
            return quantity
        else:
            raise Exception('unknown units for quantity "%s": %s'%(quantity,model_unit))
    
    pes_source = \
    'function pes(au,ido,r1,r2,r3,mass)' + '\n\n' + \
    '' + \
    'implicit none' + '\n\n' + \
    '' + \
    'logical au' + '\n' + \
    'integer ido' + '\n' + \
    'real(8) pes,r1,r2,r3,mass(*)' + '\n' + \
    'real(8) r1_,r2_,r3_' + '\n\n' + \
    '' + \
    'real(8) mass1,mass2' + '\n' + \
    'real(8) BOHR_TO_ANGSTROEM,PI_IN_RADIANS,DEGREE_TO_RADIAN,WAVENUMBER_TO_HARTREE' + '\n\n' + \
    '' + \
    'intent(in) :: au,ido,r1,r2,r3' + '\n\n' + \
    '' + \
    'parameter(BOHR_TO_ANGSTROEM=0.529177249d0)' + '\n' + \
    'parameter(PI_IN_RADIANS=3.14159265358979d0)' + '\n' + \
    'parameter(DEGREE_TO_RADIAN=1.74532925199433d-2)' + '\n' + \
    'parameter(WAVENUMBER_TO_HARTREE=0.4556335d-5)' + '\n\n' + \
    '' + \
    '!===========================================================' + '\n' + \
    '! Interface routine between VTET and PES codes' + '\n' + \
    '!===========================================================' + '\n' + \
    '' + \
    'real(8) pes1' + '\n\n' + \
    '' + \
    '! UNITS1: Convert input units according to the old pes_ expectations.' + '\n' + \
    '! au==TRUE => r in bohr, theta in rad' + '\n' + \
    '! au==FALSE => r in angstroem, rho in rad' + '\n' + \
    'if (au) then' + '\n' + \
    '    r1_ = r1' + '\n' + \
    '    r2_ = r2' + '\n' + \
    '    r3_ = r3' + '\n' + \
    'else' + '\n' + \
    '    r1_ = r1/BOHR_TO_ANGSTROEM' + '\n' + \
    '    r2_ = r2/BOHR_TO_ANGSTROEM' + '\n' + \
    '    r3_ = PI_IN_RADIANS-r3' + '\n' + \
    'endif' + '\n\n' + \
    '' + \
    '! UNITS2: Convert input units according to the generated Python model.' + '\n' + \
    'r1_ = %s'%convert_units_pes_to_model('r1_',input_units[input_names[0]]) + '\n' + \
    'r2_ = %s'%convert_units_pes_to_model('r2_',input_units[input_names[1]]) + '\n' + \
    'r3_ = %s'%convert_units_pes_to_model('r3_',input_units[input_names[2]]) + '\n\n' + \
    '' + \
    'mass1 = mass(1)' + '\n' + \
    'mass2 = mass(2)' + '\n\n' + \
    '' + \
    'call sympy_generated(pes1,r1_,r2_,r3_%s)'%(',mass1,mass2' if len(input_names)==5 else '') + '\n\n' + \
    '' + \
    '! UNITS3: Convert output units according to the old pes_ expectations.' + '\n' + \
    'pes1 = %s'%convert_units_pes_to_model('pes1',output_units) + '\n\n' + \
    '' + \
    '! UNITS4: Convert input units according to the generated Python model.' + '\n' + \
    'if (au) then' + '\n' + \
    '    pes1 = pes1*WAVENUMBER_TO_HARTREE' + '\n' + \
    'endif' + '\n\n' + \
    '' + \
    'if(ido==0) then' + '\n' + \
    '    pes=pes1' + '\n' + \
    'else' + '\n' + \
    '    stop "ERROR: noadifor version does not imply PES differentiation"' + '\n' + \
    'endif' + '\n\n' + \
    '' + \
    'end'
    
    with open('pes.f90','w') as fil:
        fil.write(pes_source)

def compile_gridcalc_fortran(model,compiler='gfortran'):
    """
    Compile the gridcalc with a model given in filename.
    The compiler executable must be accessible!
    """

    filename = model.__f90_filename__
    params_list = ','.join(model.__input_names__)
    exefile = model.__class__.__name__

    nparams = len(params_list)

    # Define the gridcalc function to return.
    def gridcalc(exefile,grid):
        # create grid file
        meshes = grid.get_meshes(flat=True)
        #meshes = [mesh.flatten() for mesh in meshes]
        gridfile = '~.pts'
        #np.savetxt(gridfile,list(zip(r1_,r2_,alpha_)))
        np.savetxt(gridfile,list(zip(*meshes)))
        # call to gridcalc program
        #subprocess.run(['./'+exefile,gridfile])
        subprocess_run_crossplatform([exefile,gridfile])
        outputfile = gridfile+'.out'
        data = np.loadtxt(outputfile)
        data = list(zip(*data))
        res = np.reshape(data[-1],grid.__shape__)
        return res
        
    # Create the gridcalc main source file.
    gridcalc_main_source = \
        'program gridcalc' + '\n\n' + \
        '!use ifport' + '\n\n' + \
        'implicit none' + '\n\n' + \
        'character gridfilename*100' + '\n' + \
        'integer ios' + '\n' + \
        'integer ch_in,ch_out' + '\n' + \
        'double precision :: %s,res'%params_list + '\n\n' + \
        'call getarg(1,gridfilename)' + '\n\n' + \
        'print *,\'gridfilename \',gridfilename' + '\n\n' + \
        'open(unit=1,file=gridfilename,action=\'READ\',status=\'OLD\')' + '\n' + \
        'open(unit=2,file=trim(gridfilename)//\'.out\',action=\'WRITE\',status=\'REPLACE\')' + '\n\n' + \
        'print *,\'start loop...\'' + '\n\n' + \
        'do while (.true.)' + '\n' + \
        '  read(1,*,iostat=ios) %s'%params_list + '\n' + \
        '  if( ios > 0 ) then' + '\n' + \
        '    stop "problem somewhere"' + '\n' + \
        '  else if( ios < 0 ) then' + '\n' + \
        '    exit' + '\n' + \
        '  endif' + '\n' + \
        '  call sympy_generated(res,%s)'%params_list + '\n' + \
        '  write(2,\'(%dE25.15)\') %s,res'%(nparams+1,params_list) + '\n' + \
        'end do' + '\n' + \
        '200 continue' + '\n' + \
        'print *,\'end loop\'' + '\n\n' + \
        'close(1)' + '\n' + \
        'close(2)' + '\n\n' + \
        'end program gridcalc'
        
    with open('gridcalc.f90','w') as fil:
        fil.write(gridcalc_main_source)  
        
    # Compile the existing Fortran source.
    subprocess.run([compiler,'-o',exefile,filename,'gridcalc.f90'])
    #subprocess_run_crossplatform([compiler,'-o',exefile,filename,'gridcalc.f90'])
    
    # Get exefile and create the resulting function.
    filestem,_ = os.path.splitext(filename)
    foo = lambda grid: gridcalc(exefile,grid)
    
    return foo
    
        
# ATTENTION =================================
# TOO MANY PARAMETERS WILL SPAWN ERROR IN LAMBDIFY:
# SyntaxError: more than 255 arguments
# POSSIBLE SOLUTION IS HERE:
# https://stackoverflow.com/questions/45220567/sympy-lambdify-function-with-large-array-input    
    
class ModelSympy(Model):
    """
    Abstract class for the fit model based on Sympy symbolic library.
    Made for two major reasons, mostly for being able to:
        1) ...compute Jacobian efficiently
        2) ...export the model and its parameters to the Fortran module
    The following variables should be set in the __init__ method:
        self.__params__: python parameter objects to make connection with the Python model
    """
    ##########################################################
    ################# CALCULATE FUNCTION #####################
    ##########################################################
    
    def __sympy_initialize_func__(self):
        print('')
        print('==========================================')
        print('<<<< calling __sympy_initialize_func__ >>>>')
        print('==========================================')
        print('Progress:')
        
        # Create Sympy objects for inputs.
        print('     - creating sympy objects for inputs')
        argnames = get_argnames(self.__func__); argnames = list(argnames)
        argnames = argnames[2:] # omit "self" and "params" arguments
        inputs_ = [sy.Symbol(argname,real=True) for argname in argnames]
        self.__symbolic_inputs__ = inputs_
        self.__input_names__ = argnames
        
        # Create Sympy objects for parameters.
        print('     - creating sympy objects for parameters')
        #sympy_pars = [sy.Symbol(p.__name__,real=True) for p in self.__params__ if p.__flag__] # ONLY ACTIVE PARAMETERS
        sympy_pars = [sy.Symbol(p.__name__,real=True) for p in self.__params__]
        params_ = copy.deepcopy(self.__params__)
        #params_.set_values(sympy_pars,active_only=True) # ONLY ACTIVE PARAMETERS
        params_.set_values(sympy_pars,active_only=False)
        self.__symbolic_params__ = params_
        
        # Get the Sympy expression by calling function with the Sympy objects.
        print('     - get the Sympy expression by calling function with the Sympy objects')
        self.__symbolic_func__ = self.__func__(params_,*inputs_)
        
        # Create lambdified Python function from sympy expression.
        print('     - create lambdified Python function from sympy expression')
        args_lambdify = [psym.get_value() for psym in self.__symbolic_params__]+self.__symbolic_inputs__
        self.__lambdified_func__ = sy.lambdify(args_lambdify,self.__symbolic_func__)
        
        # Create compiled (numbified) code from the lambdified function.
        print('     - create compiled (numbified) code from the lambdified function')
        self.__numbified_func__ = nb.njit(self.__lambdified_func__)
        
        print('==========================================\n')
                
    def __calc_symbolic__(self,params,*inputs): # symbolic expression->numeric result (through substitution) ONLY FOR DEBUGING PURPOSES, VERY SLOW!!!
        dct = {psym.get_value():params[psym.__name__].get_value() for psym in self.__symbolic_params__} # parameters substitution
        dct.update({v:inp for v,inp in zip(self.__symbolic_inputs__,inputs)})
        expr_sub = self.__symbolic_func__.subs(dct) # substitute parameters and inputs        
        return expr_sub.evalf()

    def __calc_lambdified__(self,params,*inputs): # lambdified expression->numeric result
        args = [p.get_value() for p in params]+list(inputs)
        return self.__lambdified_func__(*args)

    def __calc_numbified__(self,params,*inputs): # compiled lambdified expression->numeric result
        args = [p.get_value() for p in params]+list(inputs)
        return self.__numbified_func__(*args)
        
    def calculate_components(self,grid,compnames=[]): # calculate model and it's components on grid
        def func_comp(*x):
            # get the symbolic components
            res = self.__func__(self.__params__,*x)
            comps = [self.__components__[name] for name in compnames]
            # evaluate all components
            #res = res.evalf()   # WORKS FINE WITHOUT THIS??????
            #comps = [comp.evalf() for comp in comps]   # WORKS FINE WITHOUT THIS??????
            # return evaluated components
            return [res,*comps]
        hypermesh = grid.calculate(func_comp,dtype=object)
        meshes = [np.empty(hypermesh.shape,dtype=np.float64) for _ in range(len(compnames)+1)]
        for k,_ in enumerate(meshes):
            for i,_ in np.ndenumerate(hypermesh):
                meshes[k][i] = hypermesh[i][k]
        return meshes
        
    def __calc__(self,params,*inputs):
        # check for the model initialization
        if '__initialized_func__' not in self.__dict__ or self.__initialized_func__ is False:
            self.__sympy_initialize_func__()
            self.__initialized_func__ = True
            
        # check the validity of algorithm on each N_th calculation
        if '__calc_counter__' not in self.__dict__:
            self.__calc_counter__ = 0
        else:
            self.__calc_counter__ += 1
        N_CHECK = 100000 # have it sufficiently large to avoid littering the output
        #DO_CHECK = True
        if '__check_symbolic__' not in self.__dict__:
            DO_CHECK = True
        else:
            DO_CHECK = self.__check_symbolic__
        if DO_CHECK and self.__calc_counter__%N_CHECK==0:
            print('CALL_COUNTER: %d'%self.__calc_counter__)
            self.__check__(params,*inputs)
        
        # call on of the calc functions depending on an internal switch
        if self.__calc_switch__ == 'symbolic':
            return self.__calc_symbolic__(params,*inputs)
        elif self.__calc_switch__ == 'lambdified':
            return self.__calc_lambdified__(params,*inputs)
        elif self.__calc_switch__ == 'numbified':
            return self.__calc_numbified__(params,*inputs)
        else:
            raise Exception('unknown calc option "%s"'%self.__calc_switch__)

    ##########################################################
    ################# CALCULATE JACOBIAN #####################
    ##########################################################
        
    def __sympy_initialize_jac__(self):
        print('')
        print('==========================================')
        print('<<<< calling __sympy_initialize_jac__ >>>>')
        print('==========================================')
        print('Progress:')
                
        # Get the Sympy expression for jacobian from __symbolic_func__.
        print('     - get the Sympy expression by calling function with the Sympy objects')
        #sympy_pars = [spar.get_value() for spar in self.__symbolic_params__] # ALL PARAMETERS (CAN BE INEFFICIENT IF THE ACTIVE SUBSET IS MUCH SMALLER THAT THE TOTAL PARAMETER SET)
        sympy_pars = [spar.get_value() for spar in self.__symbolic_params__ if spar.__flag__] # !!! ONLY ACTIVE PARAMETERS (CAN CAUSE ERRORS IF THE ACTIVE SUBSET CHANGES WITHIN ONE SESSION)
        self.__symbolic_jac__ = gradient(self.__symbolic_func__,sympy_pars) # symbolic jacobian with respect to all parameters
        
        # Create lambdified Python function from sympy expression.
        print('     - create lambdified Python function from sympy expression')
        args_lambdify = [psym.get_value() for psym in self.__symbolic_params__]+self.__symbolic_inputs__
        self.__lambdified_jac__ = sy.lambdify(args_lambdify,self.__symbolic_jac__)
        
        
        # Create compiled (numbified) code from the lambdified function.
        print('     - create compiled (numbified) code from the lambdified function')
        self.__numbified_jac__ = nb.njit(self.__lambdified_jac__)
        
        print('==========================================\n')
        
    def __jac_symbolic__(self,params,*inputs): # symbolic expression->numeric result (through substitution) ONLY FOR DEBUGING PURPOSES, VERY SLOW!!!
        dct = {psym.get_value():params[psym.__name__].get_value() for psym in self.__symbolic_params__} # parameters substitution
        dct.update({v:inp for v,inp in zip(self.__symbolic_inputs__,inputs)})
        expr_sub = self.__symbolic_jac__.subs(dct) # substitute parameters and inputs  
        return np.array(expr_sub.tolist()[0],dtype=np.float64) # using tolist() method; is there a better way to do this?

    def __jac_lambdified__(self,params,*inputs): # lambdified expression->numeric result
        args = [p.get_value() for p in params]+list(inputs)
        return self.__lambdified_jac__(*args)[0]

    def __jac_numbified__(self,params,*inputs): # compiled lambdified expression->numeric result
        args = [p.get_value() for p in params]+list(inputs)
        return self.__numbified_jac__(*args)[0]
        
    def __jac__(self,params,*inputs): 
        # check for the jacobian initialization
        if '__initialized_jac__' not in self.__dict__ or self.__initialized_jac__ is False:
            self.__sympy_initialize_jac__()
            self.__initialized_jac__ = True
            
        # call on of the calc functions depending on an internal switch
        if self.__calc_switch__ == 'symbolic':
            return self.__jac_symbolic__(params,*inputs)
        elif self.__calc_switch__ == 'lambdified':
            return self.__jac_lambdified__(params,*inputs)
        elif self.__calc_switch__ == 'numbified':
            return self.__jac_numbified__(params,*inputs)
        else:
            raise Exception('unknown calc option "%s"'%self.__calc_switch__)            
            
    ##########################################################
    ############# CHECK FUNCTION CALCULATION  ################
    ##########################################################
        
    def __check__(self,params,*inputs):
        """
        Check the approximate equality of results for four methods of
        the function computations:
        1) Via __symbolic_func__: pure symbolic calculation though substitution
        2) Via __lambdified_func__: pure python function derived from __symbolic_func__
        3) Via __numbified_func__: __symbolic_func__ compiled with Numba
        4) Via __func__: initial function (using numerical parameters and inputs and symbolic functions)
        """
        print('')
        print('====================================================')
        print('<<<< checking the validity of symbolic approach >>>>')
        print('====================================================')
        print(' 1) Via __symbolic_func__: pure symbolic calculation though substitution')
        print(' 2) Via __lambdified_func__: pure python function derived from __symbolic_func__')
        print(' 3) Via __numbified_func__: __lambdified_func__ compiled with Numba')
        print(' 4) Via __func__: initial function (using numerical parameters and inputs and symbolic functions)')
        print('')
        res__symbolic_func__ = self.__calc_symbolic__(params,*inputs)
        res__lambdified_func__ = self.__calc_lambdified__(params,*inputs)
        res__numbified_func__ = self.__calc_numbified__(params,*inputs)
        res__func__ = self.__func__(params,*inputs)
        vals = [res__symbolic_func__, res__lambdified_func__, res__numbified_func__, res__func__]
        n = len(vals)
        diffs = np.zeros([n,n])
        abs_errors = []
        for i in range(n):
            val_i = vals[i]
            for j in range(i,n):
                val_j = vals[j]
                diff = val_i - val_j
                diffs[i,j] = diff
                abs_errors.append(np.abs(diff))
        min_ae = min(abs_errors); max_ae = max(abs_errors)
        print('MAX_ABS_ERROR: %f; MIN_ABS_ERROR: %f\n'%(min_ae,max_ae))
        print('RESIDUALS (i-j):')
        print(tab([[i]+list(line) for i,line in enumerate(diffs)],headers=range(n+1)))
        #THRESH = 1.0E-10
        THRESH = 1.0E-10
        QUIT_ON_ERROR = False
        if max_ae>THRESH:
            msg = 'failed with absolute error maximum of %e (greater than %e)'%(max_ae,THRESH)
            if QUIT_ON_ERROR:                
                raise Exception(msg)
            else:
                warn(msg)
        else:
            print('...PASSED VALIDITY TEST')
        print('====================================================\n')

    ##########################################################
    ############### FORTRAN CODE GENERATION  #################
    ##########################################################
            
    def generate_fortran_90(self):    # BACKUP
        """Generate Fortran 90 code from the model with current parameters"""
        subroutine_name = 'sympy_generated'
        result_var = 'calc__result__'
        
        # Create the Sympy expression
        dct = {psym.get_value():self.__params__[psym.__name__].get_value() for psym in self.__symbolic_params__} # parameters substitution
        expr_sub = self.__symbolic_func__.subs(dct) # substitute parameters and inputs        
        expr_sub_f90 = sy.fcode(expr_sub.evalf(),assign_to=result_var,standard=90,source_format='free')  # get rid of functions of constants;  

        # Create the raw Fortran code
        code = \
        'subroutine %s(%s)\n\n'%(subroutine_name,','.join([result_var]+self.__input_names__))+\
        ''.join(['double precision :: %s\n'%inpname for inpname in self.__input_names__])+\
        'double precision :: %s\n\n'%result_var+\
        expr_sub_f90+'\n\n'+\
        'end subroutine %s'%subroutine_name
        
        # Save the code to file.
        f90_filename = self.__class__.__name__+'.f90'
        print('\n==========================')
        print('CREATING THE FORTRAN CODE')
        print('==========================')
        print('   saving to file "%s"'%f90_filename)
        with open(f90_filename,'w') as fout:
            fout.write(code)
        
        # Save filename
        self.__f90_filename__ = f90_filename
        
        # Compile the grid function
        #self.__function_fortran__ = compile_gridcalc_fortran(self,f90_filename)
        
    def compile_fortran_90(self,compiler='gfortran'):
        self.__function_fortran__ = compile_gridcalc_fortran(self,compiler)
        
        #return code

#    def generate_fortran_90_with_vars(self): # CAN BE DELETED?
#        """Generate Fortran 90 code from the model with current parameters WITH ADDITIONAL VARIABLE DECLARATIONS"""
#        #subroutine_name = self.__class__.__name__
#        subroutine_name = 'sympy_generated'
#        #result_var = subroutine_name+'__result__'
#        result_var = 'calc__result__'
#        expr = self.__symbolic_func__ # original expression        
#        #expr_f90 = sy.fcode(expr,assign_to=result_var,standard=90,source_format='free')        
#        expr_f90 = sy.fcode(expr.evalf(),assign_to=result_var,standard=90,source_format='free')   # get rid of functions of constants; 
#        code = \
#        'subroutine %s(%s)\n\n'%(subroutine_name,','.join([result_var]+self.__input_names__))+\
#        ''.join(['double precision :: %s\n'%inpname for inpname in self.__input_names__])+\
#        'double precision :: %s\n\n'%result_var+\
#        ''.join(['double precision :: %s=%s\n'%\
#            (par.__name__,('%25.17E'%par.__value__).replace('E','D')) for par in self.__params__])+\
#        '\n'+expr_f90+'\n\n'+\
#        'end subroutine %s'%subroutine_name
#        self.__function_fortran__ = compile_gridcalc_fortran(self,expr_sub_f90)
#        raise NotImplementedError # create a stub for this method
#        #return code
        
# ===============================================================================
# ================================ MODEL SYMPY ==================================
# ===============================================================================



import numpy as np
from numpy.linalg import pinv, inv

# ===============================================================================
# =========================== FIT AND PENALTY POINTS ============================
# ===============================================================================

class FitGroups:
    """
    Class to implement the groups of the fit points (penalty, outputs, etc...).
    """
    def __init__(self,grps):
        """grps - list; index - dict"""
        self.__grps__ = grps
        self.__index__ = {grp.__name__:i for i,grp in enumerate(grps)}
        
    def __getitem__(self,name):
        i = self.__index__[name]
        grp = self.__grps__[i]
        return grp
        
    def __len__(self):
        return len(self.__grps__)
        
    def __iter__(self):
        return iter(self.__grps__)        

    def add_group(self,grp): 
        self.__index__[grp.__name__] = len(self.__grps__)
        self.__grps__.append(grp)
    
    def print_stat(self):
        sse_tot = 0
        for grp in self.__grps__:
            if not grp.__active__: continue
            weighted_resids = grp.__weighted_resids__
            unweighted_resids = grp.__unweighted_resids__
            calc_vals = grp.__calc_vals__
            wht_sse = np.sum(weighted_resids**2)
            unwht_sse = np.sum(unweighted_resids**2)
            whts = grp.__whtmul__*grp.__weights__
            min_wht = np.min(whts); max_wht = np.max(whts)
            wht_sd = np.sqrt(wht_sse/grp.length)
            unwht_sd = np.sqrt(unwht_sse/grp.length)
            N = np.prod(weighted_resids.shape)
            print('%30s:   N:%5d   MIN_WHT: %5.1e   MAX_WHT: %5.1e   WHT_SSE: %7.3e   UNWHT_SSE: %7.3e   WHT_SD: %7.3e   UNWHT_SD: %7.3e'%\
            (grp.__name__,N,min_wht,max_wht,wht_sse,unwht_sse,wht_sd,unwht_sd))
            
#            sse_tot += wht_sse
#        print('datagroups->sse_tot>>>','%.9e'%sse_tot)
            
    def split_global_jacobian(self,jac):
        """
        Split "global" Jacobian matrix between the active groups.
        Needed for calculation of the fit statistics for each group separately.
        """        
        offset = 0
        for grp in self.__grps__:
            if not grp.__active__: continue
            grp.__jacobian__ = jac[offset:(offset+grp.length)]
            offset += grp.length
            
    #def __repr__(self):
    #    return self.__name__

#class AbstractFitPoints(ABC):
class AbstractFitPoints:
    """
    Abstract class for Fitting points.
    Has the grids for input (N per M) and desired output (N per 1) points,
    where N is number of "observables", M is number of dimensions.
    Children must implement the residuals and their derivatives for 
    the given output values.
    """
    def __init__(self,name,input,output,whtmul=1,whts=None,active=True):
        self.__name__ = name
        self.__active__ = active
        self.__inputgrid__ = input
        self.__output__ = np.array(output)
        self.__whtmul__ = whtmul
        if whts is not None:
            self.__weights__ = np.array(whts)
        else:
            self.__weights__ = np.full(output.shape,fill_value=1.0)
        self.__calc_vals__ = None
        self.__weighted_resids__ = None
        self.__unweighted_resids__ = None
        
    def __getitem__(self,varname):
        """Get the mesh of the corresponding __inputgrid__ by the variable name"""
        return self.__inputgrid__[varname]
            
    @property
    def length(self):
        return self.__output__.size        
        
    def set_weights(self,whts):
        self.__weights__ = np.array(whts)
        
    def set_calc_vals(self,calc_vals):
        """Set last calculated values, and calculate 
           weighted and unweighted residuals"""
        self.__calc_vals__ = calc_vals
                        
    def calculate_residuals(self):
        """Array of residuals between output and calc_vals"""        
        self.__unweighted_resids__ = np.zeros(self.__output__.shape)
        # calculate unweighted resids
        for i,_ in np.ndenumerate(self.__output__):
            self.__unweighted_resids__[i] = self.resid_func(self.__output__[i],self.__calc_vals__[i])
        # calculate weighted resids
        self.__weighted_resids__ = self.__whtmul__*self.__weights__*self.__unweighted_resids__
        
    def calculate_residual_derivs(self):
        """Array of derivatives of residuals by 'calc'"""
        self.__unweighted_resid_derivs__ = np.zeros(self.__output__.shape)
        # calculate unweighted resids
        for i,_ in np.ndenumerate(self.__output__):
            self.__unweighted_resid_derivs__[i] = self.resid_deriv_func(self.__output__[i],self.__calc_vals__[i])
        # calculate weighted resids
        self.__weighted_resid_derivs__ = self.__whtmul__*self.__weights__*self.__unweighted_resid_derivs__   
        
    def stats(self,statname):
        """
        Useful shortcut to the fit statistics of all types.
        """
        if statname=='cook':
            return self.__cook__
        elif statname=='dffits':
            return self.__DFFITS__
        elif statname=='leverage':
            return self.__leverage__
        elif statname=='student':
            return self.__weighted_studentized_resids__
        else:
            raise Exception('unknown statistic %s'%statname)
        
    def calculate_fit_statistics(self):
        """Calculate useful fit statistics (cook, jackknife etc...).
           For some of the statistics Jacobian matrix is required.
           
           Used the following references:
           https://en.wikipedia.org/wiki/Leverage_(statistics)
           https://en.wikipedia.org/wiki/Studentized_residual
           https://en.wikipedia.org/wiki/DFFITS
           https://www.mathworks.com/help/stats/cooks-distance.html
           """
        X = self.__jacobian__   
        n = X.shape[0] # number of observables
        m = X.shape[1] # number of free parameters
        eps = self.__weighted_resids__; eps = eps.flatten() # !!!! IF FLATTEN IS OMITTED, THEN DIMENSIONALITY MISMATCH OCCURS; ANOTHER WAY OF SOLVING THIS IS TO KEEP THE ORIGINAL DIMENSIONALITY (TODO!!!!)
        eps2 = eps**2
        sse = np.sum(eps2) # sum of squares error
        mse = sse/n # mean squared error
           
        # Calculate Hat matrix
        H = X @ pinv(X.T @ X) @ X.T # !!! using generalized inversion (pinv) since rank(X)<n_p
        
        # Leverages
        leverage = np.diag(H)
        self.__leverage__ = leverage
        
        # Studentized residuals (weighted)
        sigma2 = 1/(n-m)*sse
        T = eps/(np.sqrt(sigma2)*np.sqrt(1-leverage))
        self.__weighted_studentized_resids__ = T
        
        # DFFITS
        DFFITS = T*np.sqrt(leverage/(1-leverage))
        self.__DFFITS__ = DFFITS
        
        # Cook's distances
        cook = eps2/(m*mse)*(leverage/(1-leverage)**2)
        self.__cook__ = cook        

    #@abstractmethod
    def resid_func(self,obs,calc):
        """Residual, scalar"""
        raise NotImplementedError

    #@abstractmethod
    def resid_deriv_func(self,obs,calc):
        """Derivative by calc, scalar"""
        raise NotImplementedError
        
    def __repr__(self):
        return self.__name__
        
class FitPoints(AbstractFitPoints): # can be sped up by re-defining resids and d_resids
    def resid_func(self,obs,calc):
        return obs-calc
    
    def resid_deriv_func(self,obs,calc):
        return -1
    
class PenaltyPointsUp(AbstractFitPoints):
    def resid_func(self,obs,calc):
        return obs-calc if obs<calc else 0
        
    def resid_deriv_func(self,obs,calc):
        return -1 if obs<calc else 0

class PenaltyPointsDown(AbstractFitPoints):
    def resid_func(self,obs,calc):
        return obs-calc if obs>calc else 0
        
    def resid_deriv_func(self,obs,calc):
        return -1 if obs>calc else 0



""" New version of the fit Parameters based on Jeanny collections."""

import numpy as np
from tabulate import tabulate as tab

from jeanny3 import Collection

class Parameters(Collection):
    def __init__(self,N=0,prefix=None,names=None,
                 values=None,flags=True,mins=None,maxs=None,
                 weights=None,group='default'):
        self.initialize() # call initialization of a Collection
        #self.__index__ = {}
        self.generate(N=N,prefix=prefix,group=group,
                      names=names,values=values,flags=flags,
                      mins=mins,maxs=maxs,weights=weights)
        self.order = ['group','name','value','min','max','weight','flag','__id__']
                                    
    def __iter__(self):
        return iter(self.getitems()) # strange peculiarity of Python 3
        
    def __len__(self):
        return len(self.ids())
                      
    @property
    def __npar__(self):
        return len(self.ids())
        
    #def __getitem__(self,name):
    #    """
    #    Get parameter object by its name.
    #    """
    #    return self.__dicthash__[self.__index__[name]]
    def __getitem__(self,ID):
        """
        Get parameter object by its name.
        ID = name
        """
        return self.__dicthash__[ID]
        
    def append(self,pars,group='default'):
        """
        Append a parameter to the set. Pars can be either a parameter,
        or a list of parameters.
        """
        if type(pars) is Par:
            pars = [pars]
        elif type(pars) not in {list,tuple}:
            raise Exception('Either Par, or list/tuple of Pars are allowed')
        for par in pars:
            par['group'] = group
            name = par['name']
            #ID = (group,name)
            ID = name
            if ID in self.__dicthash__:
                raise Exception('%s already in __dicthash__ '
                    '(consider renaming parameter)'%str(ID))
            self.__dicthash__[ID] = par
                                       
    def generate(self,N=None,prefix=None,names=None,group='default',
                 values=None,flags=True,mins=None,maxs=None,weights=None):
        """
        Generate n parameters with predefined values.
        """
        if N is None and names is None:
            raise Exception('both N and names are None')
        if prefix is None: prefix = ''
        if names is None:
            # generate names:
            names = [prefix+'%d'%i for i in range(N)]
        else:
            if len(names)!=N:
                raise Exception('len(names)!=N')
        N = len(names)
        if type(weights) not in {list,tuple}:
            weights = [weights for i in range(N)]
        if type(values) not in {list,tuple}:
            values = [values for i in range(N)]
        if type(flags) not in {list,tuple}:
            flags = [flags for i in range(N)]
        if type(mins) not in {list,tuple}:
            mins = [mins for i in range(N)]
        if type(maxs) not in {list,tuple}:
            maxs = [maxs for i in range(N)]
        for name,weight,value,flag,min,max in \
                zip(names,weights,values,flags,mins,maxs):
            self.append(Par(name=name,value=value,flag=flag,
                    min=min,max=max,weight=weight),
                    group=group)

    def get_groups(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array(self.getcol('group',IDs=ii))
                    
    def get_names(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array(self.getcol('name',IDs=ii))
        
    def get_values(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array(self.getcol('value',IDs=ii))

    def set_values(self,values,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)        
        for i,val in zip(ii,values):
            self.__dicthash__[i]['value'] = val
        
    def get_bounds(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        data = self.getcols(['min','max'],IDs=ii)
        return list(zip(*data))
        
    def get_weights(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array(self.getcol('weight',IDs=ii))
    
    def get_flags(self,active_only=False,group=None):
        ii = self.get_index(active_only=active_only,group=group)
        return np.array(self.getcol('flag',IDs=ii))
               
    def get_index(self,active_only=False,group=None):   
        cond_group = (lambda v: v['group']==group) if group else (lambda v: True)
        cond_active = (lambda v: v['flag']) if active_only else (lambda v: True)
        cond = lambda v: cond_group(v) and cond_active(v)
        return self.ids(cond)
        
    def __repr__(self):
        groups = self.get_groups()
        names = self.get_names()
        values = self.get_values()
        bounds = self.get_bounds()
        weights = self.get_weights()
        flags = self.get_flags()
        indices = []; i = 0
        for flag in flags:
            if flag:
                indices.append(i); i += 1
            else:
                indices.append('')
        return tab(zip(indices,groups,names,values,bounds,weights,flags),
                headers=('#','group','names','values','bounds','weights','flags'))

def dumpval(val):
    """ Simple alternative to json.dumps for VERY limited set of value types:
        1) only scalar types are supported: int, float, str, bool
    """
    if type(val) is str:
        return '"%s"'%val
    else:
        return str(val)
                
class Par(dict):
    """
    Fitting parameter storing information about initial values, bounds, rubber weight, and active flag.
    Each parameter has its unique name.
    """
    
    def __init__(self,name,value=None,flag=None,min=None,max=None,weight=None,group=None):
        self['name'] = name
        self['group'] = 'default' if group is None else group
        self['value'] = np.random.random() if value is None else value
        self['flag'] = False if flag is None else flag
        self['min'] = -np.inf if min is None else min
        self['max'] = np.inf if max is None else max
        self['weight'] = 0.0 if weight is None else weight
    
    @property
    def __flag__(self):
        return self['flag']
    
    @property
    def __name__(self):
        return self['name']
    
    def get_value(self):
        return self['value']
        
    def set_value(self,value):
        self['value'] = value
        
    def get_flag(self):
        return self['flag']
        
    def set_flag(self,flag):
        self['flag'] = flag        
        
    def get_bounds(self):
        return [self['min'],self['max']]
        
    def set_bounds(self,bnds):
        self['min'] = bnds[0]
        self['max'] = bnds[1]
        
    def get_min(self):
        return self['min']
        
    def set_min(self,min):
        self['min'] = min

    def get_max(self):
        return self['max']
        
    def set_max(self,max):
        self['max'] = max
        
    def get_weight(self):
        return self['weight']
        
    def set_weight(self,wht):
        self['weight'] = wht
        
    def get_name(self):
        return self['name']    
        
    def get_group(self):
        return self['group']
        
    def __repr__(self):
        group = self.get_group()
        name = self.get_name()
        flag = self.get_flag()
        value = self.get_value()
        min = self.get_min()
        max = self.get_max()
        weight = self.get_weight()
        return '{}:{}:{}({}/[{}:{}])*{}'.format(
            group,name,flag,value,min,max,weight)



import numpy as np

def assert_(desc,vars):
    assert set(vars) not in set(desc.keys()),\
            'name mismatch'

def flatten(*args):
    if len(args)==1:
        return args[0].flatten()
    else:
        return [_.flatten() for _ in args]
        
class Grid:
    """
    desc: {'v1':grid1, 'v2':grid2 ...}
    vars: ['v1','v2',...]
    """
    def __init__(self,*argc):
        desc = dict([(argc[i],argc[i+1]) for i in range(0,len(argc),2)]); vars = argc[0::2]
        assert_(desc,vars)
        self.__desc__ = {_:np.array(desc[_]) for _ in desc}
        self.__vars__ = vars
        grids = [self.__desc__[name] for name in vars]        
        meshes = np.meshgrid(*grids)
        self.__meshes__ = meshes
        self.__size__ = meshes[0].size
        self.__shape__ = meshes[0].shape
        self.__hash__ = {name:i for i,name in enumerate(vars)} # hash for lookup mesh by the variable name
    
    def __getitem__(self,varname):
        return self.__meshes__[self.__hash__[varname]]
        
    def flatten(self):
        return flatten(*self.__meshes__)
        
    def get_meshes(self,ind=None,flat=False,squeeze=True):
        # https://docs.scipy.org/doc/numpy-1.15.1/reference/generated/numpy.squeeze.html
        meshes = self.__meshes__
        if flat:
            meshes = flatten(*meshes)
        if squeeze:
            meshes = [np.squeeze(_) for _ in meshes]
        if len(meshes)==1: meshes=meshes[0]
        if ind is None:
            return meshes
        else:
            return [mesh[ind] for mesh in meshes]
                                    
    def calculate(self,func,flat=False,squeeze=True,dtype=np.float64): # calculate arbitrary function on the grid; SEEMS IT CAN BE NUMBIFIED! (ndenumerate IS SUPPORTED BY NUMBA)        
        meshes = self.__meshes__
        res = np.empty(meshes[0].shape,dtype=dtype)
        for i,_ in np.ndenumerate(res): # np.ndenumerate, np.nditer
            arg = [mesh[i] for mesh in meshes]
            res[i] = func(*arg)
        if flat:
            res = flatten(res)
        if squeeze:
            res = np.squeeze(res)
        return res
        
class List:
    """
    desc: {'v1':grid1, 'v2':grid2 ...}
    vars: ['v1','v2',...]
    """
    def __init__(self,*argc):
        desc = dict([(argc[i],argc[i+1]) for i in range(0,len(argc),2)]); vars = argc[0::2]
        assert_(desc,vars)
        self.__desc__ = {_:np.array(desc[_]) for _ in desc}
        self.__vars__ = vars
        meshes = [self.__desc__[name] for name in vars]
        self.__meshes__ = meshes
        self.__size__ = meshes[0].size
        self.__shape__ = meshes[0].shape
        self.__hash__ = {name:i for i,name in enumerate(vars)} # hash for lookup mesh by the variable name
        
    def __getitem__(self,varname):
        return self.__meshes__[self.__hash__[varname]]        
        
    def flatten(self): # newly added
        return flatten(*self.__meshes__)
                        
    def get_meshes(self,ind=None,flat=False):
        meshes = self.__meshes__
        if len(meshes)==1: meshes=meshes[0]
        if ind is None:
            return meshes
        else:
            return [mesh[ind] for mesh in meshes]
                                    
    def calculate(self,func,flat=False,dtype=np.float64): # calculate arbitrary function on the grid; SEEMS IT CAN BE NUMBIFIED! (ndenumerate IS SUPPORTED BY NUMBA)
        meshes = self.__meshes__
        res = np.empty(meshes[0].shape,dtype=dtype)
        for i,_ in np.ndenumerate(res): # np.ndenumerate, np.nditer
            arg = [mesh[i] for mesh in meshes]            
            res[i] = func(*arg)
        return res



from ..base import ConfigSection

class Default(ConfigSection):
    __name__ = 'PLOTTING'
    __header__ = 'PLOTTING OPTIONS'
    __template__ = \
"""
# ATTENTION: names of the plot coordinates correspond to the 
# model argumen names given in the MODEL section.

# Type of the plot.
# Available plot modes:
#   => "residuals": 
#           plot unweighted fit residuals.
#   => "sections": 
#           plot sections/cuts of the fitted model vs datapoints.
{plot_mode}

# Plot coordinate grid specifications.
# Format of the specification must be as follows: 
# X=XVAL; Y=YMIN:YSTEP:YMAX; ...
# In case of the fixed coordinate, variable would only have a value, e.g. X=XVAL.
# In case of unfixed coordinate, there should be a full grid, e.g. Y=YMIN:YSTEP:YMAX
# N.B.: 1) order of the unfixed coords affects the order of the plot axes.
#       2) order of the binding MUST correspond to the argument of the model's 
#          __func__ method.
{gridspec}

# Model components to plot.
{model_components}

# Calculate model's components.
{calculate_components}

# Plot outlier statistics in color. If False, each datagroup has its own color.
{plot_outlier_stats}

# Plot weighted residuals.
{resids_weighted}

# X axes to plot the resuduals versus. 
# If empty, defaults to [DATA][output_column].
{resids_x_axes}

# Scatter settings (2D, 3D case)
{scatter_opacity}
{marker_size}
{resize_by_weights}

# Surface settings (3D case)
{surface_opacity}
"""
    plot_mode = 'residuals'
    scatter_opacity = 1
    resize_by_weights = False
    surface_opacity = 0.4
    calculate_components = False
    resids_weighted = False
    plot_outlier_stats = False
    marker_size = 20



from ..base import ConfigSection

class Default(ConfigSection):
    __name__ = 'CODEGEN'
    __header__ = 'CODE GENERATOR OPTIONS'
    __template__ = \
"""
# Create Fortran code.
{create_fortran}

# Compare original Python code with the generated Fortran.
{compare_fortran}

# Fortran compiler executable
{compiler_fortran}

# Grid specifications to calculate on.
# Format of the specification must be as follows: 
# X=XMIN:XSTEP:XMAX; Y=YMIN:YSTEP:YMAX; ...
{gridspec}
"""
    create_fortran = False
    compare_fortran = False
    compiler_fortran = 'ifort'



from ..base import ConfigSection

class Default(ConfigSection):
    __name__ = 'STAT'
    __header__ = 'STATISTICS CALCULATION OPTIONS'
    __template__ = \
"""
# Fit statistics file.
{stat_file}

# Calculate outlier statistics.
{outlier_stats_flag}

# Type of statistics: cook, dffits, leverage ,student.
{outlier_stats_type}

# Output generated symbolic function, for debug purposes only.
{output_symbolic_func}
"""
    stat_file = 'stat.out'
    outlier_stats_flag = True
    outlier_stats_type = 'cook'



import dill

def serialize_fit(fit,filename):
    if '__dll_func__' in fit.__model__.__dict__:
        fit.__model__.__dll_func__ = None # remove lambda function to make Pickle work
    if '__initialized_func__' in fit.__model__.__dict__:
        fit.__model__.__initialized_func__ = False # remove __initialized_func__ to make Pickle work
    if '__initialized_jac__' in fit.__model__.__dict__:
        fit.__model__.__initialized_jac__ = False # remove __initialized_jac__ to make Pickle work
    if '__lambdified_func__' in fit.__model__.__dict__:
        fit.__model__.__lambdified_func__ = None # remove __lambdified_func__ to make Pickle work    
    if '__numbified_func__' in fit.__model__.__dict__:
        fit.__model__.__numbified_func__ = None # remove __numbified_func__ to make Pickle work    
    if '__lambdified_jac__' in fit.__model__.__dict__:
        fit.__model__.__lambdified_jac__ = None # remove __lambdified_jac__ to make Pickle work    
    if '__numbified_jac__' in fit.__model__.__dict__:
        fit.__model__.__numbified_jac__ = None # remove __numbified_jac__ to make Pickle work   
    if '__function_fortran__' in fit.__model__.__dict__:
        fit.__model__.__function_fortran__ = None # remove __function_fortran__ to make Pickle work           
    with open(filename, 'wb') as output:
        dill.dump(fit, output, dill.DEFAULT_PROTOCOL)
       
def serialize_model(model,filename):
    with open(filename,'wb') as fil:   
        if '__dll_func__' in model.__dict__:
            model.__dll_func__ = None # remove lambda function to make Pickle work
        if '__initialized_func__' in model.__dict__:
            model.__initialized_func__ = False # remove __initialized_func__ to make Pickle work
        if '__initialized_jac__' in model.__dict__:
            model.__initialized_jac__ = False # remove __initialized_jac__ to make Pickle work
        if '__lambdified_func__' in model.__dict__:
            model.__lambdified_func__ = None # remove __lambdified_func__ to make Pickle work    
        if '__numbified_func__' in model.__dict__:
            model.__numbified_func__ = None # remove __numbified_func__ to make Pickle work    
        if '__lambdified_jac__' in model.__dict__:
            model.__lambdified_jac__ = None # remove __lambdified_jac__ to make Pickle work    
        if '__numbified_jac__' in model.__dict__:
            model.__numbified_jac__ = None # remove __numbified_jac__ to make Pickle work    
        dill.dump(model,fil,dill.DEFAULT_PROTOCOL)

def deserialize_fit(filename):
    with open(filename, 'rb') as input:
        return dill.load(input)

def deserialize_model(filename):
    with open(filename, 'rb') as input:
        return dill.load(input)



from ..base import ConfigSection

class Default(ConfigSection):
    __name__ = 'CALC'
    __header__ = 'CALCULATION OPTIONS'
    __template__ = \
"""
# Output file to output calculated model. 
# Column names are defined in DATA section.
{output_file}

# Grid specifications to calculate on.
# Format of the specification must be as follows: 
# X=XMIN:XSTEP:XMAX; Y=YMIN:YSTEP:YMAX; ...
{gridspec}
"""
    output_file = 'calc.csv'


from ..base import ConfigSection

class Default(ConfigSection):
    __name__ = 'MODEL'
    __header__ = 'FITTING MODEL SPECIFICATIONS'
    __template__ = \
"""
# Model package name.
{model}

# Model arguments. Argument names must be in the same order
# as the data column names from the DATA section.
# E.g.: X;Y;Z
{arguments}
"""
    model = 'fitmodel'



import numpy as np
import jeanny3 as j
from algopy import UTPM

class Model:
    """
    Abstract class for the fit model supporting algorithmic differentiation.
    Description of methods:
        1) self.__func__: the function written in Python, accounts for params and components
        1) self.__calc__: interface between __func__ and the internal machinery
                          __calc__ should be re-defined when making an inherited class (e.g. ModelSympy) 
        3) self.__jac__: usually generates automatically from __func__ or __calc__
    """       
    @property
    def __p__(self): # array of values of active ("unfixed") parameters
        return self.__params__.get_values(active_only=True)
        
    @property
    def __wp__(self): # weights of active ("unfixed") parameters
        return self.__params__.get_weights(active_only=True)
        
    @property
    def __npar__(self):
        return self.__params__.__npar__

    def __call__(self,*inputs):
        """
        Making possible to call the model as a function (magic method).
        Don't confuse with __calc__!!!
        """
        return self.__calc__(self.__params__,*inputs)
        
    def __calc__(self,params,*inputs):
        """
        Interface between the func and Jacobian.
        Don't confuse with __call__!!!
        """
        return self.__func__(params,*inputs)
            
    def calculate(self,grid): # calculate model on grid defined in grids.py
        return grid.calculate(lambda *x:self.__calc__(self.__params__,*x))

    def calculate_components(self,grid,compnames=[]): # calculate model and it's components on grid
        def func_comp(*x):
            res = self.__calc__(self.__params__,*x)
            comps = [self.__components__[name] for name in compnames]
            return [res,*comps]
        hypermesh = grid.calculate(func_comp,dtype=object)
        meshes = [np.empty(hypermesh.shape,dtype=np.float64) for _ in range(len(compnames)+1)]
        for k,_ in enumerate(meshes):
            for i,_ in np.ndenumerate(hypermesh):
                meshes[k][i] = hypermesh[i][k]
        return meshes
        
    def calculate_jac(self,grid):# returns flattened array
        jac_tensor = grid.calculate(lambda *x:self.__jac__(self.__params__,*x),dtype=object,flat=True)
        jac_tensor = np.vstack(jac_tensor)
        return jac_tensor
                   
    def __jac__(self,params,*inputs): 
        # Using algopy direct mode        
        def foo(p):
            params_ = copy.deepcopy(params)
            params_.set_values(p,active_only=True)
            return self.__calc__(params_,*inputs)
        t = time()
        p = UTPM.init_jacobian(list(params.get_values(active_only=True))) 
        foo_val = foo(p)
        algopy_jacobian = UTPM.extract_jacobian(foo_val)
        return algopy_jacobian   
        
    #@abstractmethod
    def __func__(self,params,*inputs):
        raise NotImplementedError
        
    def __units__(self):
        """ 
            Return units of the inputs and output.
            The output must be a nested dict, e.g.:
                {   
                    'inputs': {'r1':'bohr',r2:'bohr',alpha:'deg'},
                    'output': 'cm-1'
                }
        """
        raise NotImplementedError
        
    def save_params(self,filename):
        self.__params__.export_csv(filename)
        
    def load_params(self,filename):
        col = j.import_csv(filename)
        #col.index(lambda v:(v['group'],v['name']))
        col.index(lambda v:v['name'])
        for ID in col.__dicthash__:
            #group,name = ID
            item = col.__dicthash__[ID]
            if ID in self.__params__.__dicthash__:
                self.__params__.__dicthash__[ID].update(item)
            else:
                self.__params__.__dicthash__[ID] = item
        
    def __repr__(self):
        return str(self.__params__)



from ..base import ConfigSection

class Default(ConfigSection):
    __name__ = 'DATA'
    __header__ = 'FITTING DATA SPECIFICATION'
    __template__ = \
"""
# CSV file containing fitting data.
{datafile}

# Define rules to split the datafile to dataspec.
# Should be ignored if dataspec is empty.
{split_column}
{split_values}
{split_weights}

# Resulting data specification split-file.
# N.B.: if empty, then the raw datafile is used.
# Data specification *.txt file has the following format:
# alias   path       wht_mul   type    include
# ------- ---------  --------  ------  -------
# test1   test1.csv  1         0       1       
# test2   test2.csv  1         0       1
# test3   test3.csv  1         0       1
{dataspec}

# Names and units of the input data columns. 
# Input names must be separated by semicolon 
# and should not contain dots.
# E.g.: X;Y;Z for names, X:X_UNIT;Y:Y_UNIT for units
{input_columns}
{input_units}

# Name of the output data column. See explanation above.
{output_column}
{output_units}

# Weight function.
{wht_fun}

# Global data cutoff
{global_cutoff_max}
{global_cutoff_min}
"""
    wht_fun = 'lambda v: 1'
    dataspec = 'dataspec.txt'



from .base import Model

# ===============================================================================
# ================================ MODEL FORTRAN ================================
# ===============================================================================

# ============
# COMPILERS
# ============

#class Compiler(ABC):
class Compiler:
    """
    Abstract class for compiler.
    """
    def __init__(self,files=[],options=[]):
        self.__files__ = files # wildcards can be used depending on the compiler
        self.__options__ = options # options  to the compiler 
        
    #@abstractmethod
    def run(self):
        raise NotImplementedError
        
    def set_files(self,*files):
        self.__files__ = files
        
    def set_options(self,*options):
        self.__options__ = options
        
class IFort(Compiler):
    """
    Intel Fortran compiler
    """
    def run(self):
        cmd = ' '.join(['ifort',*self.__options__,*self.__files__])
        os.system(cmd) 

# ====================
# SERIALIZABLE DATA        
# ====================
        
def printval(val,type):
    """
    Print the value using the format which 
    minimizes the loss of accuracy.
    """
    if type in {'double precision',
        'float(8)','float(16)','float(32)'}:
        return '%25.15E'%val
    else:
        return str(val)
        
def packvals(array,type,n_per_line=None,by_cols=False,comments=[]):
    """
    Pack values using comma-separated format specifying
    whether to pack by columns (by_cols=True) or by lines (by_cols=False)
    """
    assert len(comments)==0 or len(comments)==len(array),\
        'comments must be either empty or the same length as array'
    if n_per_line is None:
        if type=='integer':
            n_per_line = 20
        elif type in {'double precision','float(8)','float(16)'}:
            n_per_line = 5
        else:
            n_per_line = 3
    if by_cols:
        order = 'F'
    else:
        order = 'C'
    vals = array.flatten(order=order) # flatten using the Fortran order
    buf = ''
    comm = ''
    for i,val in enumerate(vals):
        buf += printval(val,type)
        if i!=len(vals)-1: 
            buf += ', '
        else:
            buf += '  '
        if len(comments)>0: comm += ' '+comments[i]
        #if (i+1)%n_per_line==0 and i!=len(vals)-1:
        if (i+1)%n_per_line==0:
            buf += '  &'
            if len(comments)>0: buf+=' !'+'%3d !'%(i+1)+comm
            buf += '\n'
            comm = ''
    #buf = buf[:-2] # get rid of the last comma and new line
    #buf = buf[:-1] # get rid of the last comma
    return buf
    
class Data:
    """
    Container class for the serializable data items.
    """
    def __init__(self,*items):
        items = list(items)
        self.__items__ = items
        self.__index__ = {item.__name__:i \
            for i,item in enumerate(items)}
        
    def __getitem__(self,name):
        i = self.__index__[item.__name__]
        return self.__items__[i]
        
    def append(self,item):
        if item.__name__ in self.__index__:
            i = self.__index__[item.__name__]
            self.__items__[i] = item
        else:
            self.__index__[item.__name__] = len(self.__items__)
            self.__items__.append(item)
        
    def __get_code__(self):
        code = ''
        for item in self.__items__:
            code += item.__get_code__()
        return code
        
#class DataItem(ABC):
class DataItem:
    """
    Abstract class for the data storage.
    """
    #@abstractmethod
    def __get_code__(self):
        raise NotImplementedError

class FortranValue(DataItem):
    """
    Storage for a scalar Fortran value.
    """
    def __init__(self,name,type,val):
        self.__name__ = name
        self.__fortran_type__ = type
        self.__value__ = val
            
    def __get_code__(self):
        return textwrap.dedent("""
        {type} :: {name} = {value}
        """.format(type=self.__fortran_type__,
                   name=self.__name__,
                   value=self.__value__))
        
class FortranArray(DataItem):
    """
    Storage for a 1D Fortran array
    """
    def __init__(self,name,type,lst,n_per_line=None):
        self.__name__ = name
        self.__fortran_type__ = type
        self.__array__ = np.array(lst)
        self.__n_per_line__ = n_per_line
            
    def __get_code__(self):
        return textwrap.dedent("""
        {type},dimension({n}) :: {name} = &
        """.format(type=self.__fortran_type__,
                   name=self.__name__,
                   n=self.__array__.size)) + \
        '(/&\n' + packvals(self.__array__,self.__fortran_type__,
                        n_per_line=self.__n_per_line__,by_cols=True) +\
        '/)'
        
class FortranMatrix(DataItem):
    """
    Storage for a 2D Fortran array
    """
    def __init__(self,name,type,lst,n_per_line=None):
        self.__name__ = name
        self.__fortran_type__ = type # e.g. "double precision"
        self.__matrix__ = np.array(lst)
        self.__n_per_line__ = n_per_line
    
    def __get_code__(self):
        return textwrap.dedent("""
        {type},dimension({ni}, {nj}) :: {name} = &
        """.format(type=self.__fortran_type__,
                   name=self.__name__,
                   ni=self.__matrix__.shape[0],   # CHECK THE FORTRAN ORDER OF ROWS/COLUMNS!!!
                   nj=self.__matrix__.shape[1])) + \
        '(/&\n' + packvals(self.__matrix__,self.__fortran_type__,
                        n_per_line=self.__n_per_line__,by_cols=True) +\
        '/)'
        
# ===================================================
# TEMPORARY FORTRAN CODE GENERATOR FOR PYTHON MODELS
# ===================================================
# NOTE: __func__ must be manually translated from Python to Fortran!!!
# NOTE2: This class has nothing in common with the rest of the Fortran-oriented 
#        classes defined above!!!

class FortranModule():
    """
    Generate Fortran module from the model object.
    The main function (__func__) is not included since
    it is not trivial to convert Python code to Fortran.
    It attempts to deduce the type and dimensions of the value automatically.
    LIMITATIONS:
        1) Currently, the following data types are supported: 
                int, np.int32, np.int64, 
                float, np.float64, 
                list, np.ndarray
        2) Only one- and two-dimensional arrays are supported.
    """
    def __init__(self,model,item_names):
        # Fitting models
        self.__model__ = model
        # Deduce types and dimensions
        self.__data__ = Data()
        for name in item_names:
            if name[0]=='_': 
                name_ = 'fort'+name # fix leading underscore
                print('WARNING: %s was changed to %s'%(name,name_))
            else:
                name_ = name
            # start deducing...
            val = getattr(model,name)
            if type(val) in {int,np.int32,np.int64}:
                item = FortranValue(name_,'integer',val)
            elif type(val) in {float,np.float64}:
                item = FortranValue(name_,'fouble precision',val)
            elif type(val) in {list,np.ndarray}:
                if type(val)==list: val = np.array(list)
                dtype = val.dtype
                if dtype in {np.dtype('int32'),np.dtype('int64')}: # np.dtype('int32')==np.int32=True, BUT np.dtype('int32') is np.int32=False !!!
                    typestr = 'integer'
                elif dtype in {np.dtype('float32'),np.dtype('float64')}:
                    typestr = 'double precision'
                else:
                    print('dtype>>>',dtype)
                # insert additional item storing the leading dimension of an array/matrix
                item_ = FortranValue('n_'+name_,'integer',val.shape[0])
                self.__data__.append(item_)
                # insert the array/matrix...
                if len(val.shape)==1:
                    item = FortranArray(name_,typestr+'(%d)'%val.shape[0],val)
                elif len(val.shape)==2:
                    item = FortranMatrix(name_,typestr,val.T,n_per_line=val.shape[1])
                else:
                    raise Exception('too many dimensions: ',val.shape)
            self.__data__.append(item)
                
    def generate_code(self):
        # generate code for items
        itemcode = self.__data__.__get_code__()
        
        # generate code for for parameters (__params__)
        params = self.__model__.__params__
        npar = len(params)
        parnames = params.get_names()
        parvalues = params.get_values() 
        parcode = textwrap.dedent("""
        double precision, parameter :: npar = {npar}
        double precision, dimension(npar) :: p = &
        (/&{new_line}{param_values}/)
        """).format(new_line='\n',npar=npar,param_values=\
            packvals(parvalues,type='double precision',n_per_line=1,comments=parnames))
        
        # insert additional useful constants
        constcode = textwrap.dedent("""
        double precision :: PI_IN_RADIANS=3.14159265358979d0
        double precision :: WAVENUMBER_TO_HARTREE=0.4556335d-5
        """)
        
        module_code = textwrap.dedent("""
        module {module_name}
        
        implicit none
        {itemcode}
        {parcode}
        {constcode}        
        end module {module_name}
        """).format(module_name=self.__model__.__class__.__name__,
                itemcode=itemcode,parcode=parcode,constcode=constcode)
        
        return module_code.strip()
        
# =============================
# FORTRAN MODEL IMPLEMENTATION
# =============================
    
class ModelFortran(Model):
    """
    Abstract class for the fit model based on Fortran routines.
    THe model requires Intel Fortran compiler and Adifor (for Jacobian)
    Adding another compilers is planned in the future if this model will be useful.
    The following methods are reimplemented: __func__, __jac__
    The following variables should be set in the __init__ method:
        self.__code__: the function code in a specific format
        self.__data__: supplementary data to be read by the code from the Fortran module
        self.__name__: name, needed to create files on the disc
        self.__params__: python parameter objects to make connection with the Python model
    """
    def __init__(self):
        
        raise NotImplementedError # model fortran is not yet implemeneted
        
        self.__module_f90_name__ = 'module_'+self.__name__
        self.__module_jac_f90_name__ = 'module_'+self.__name__+'__adifor'
        self.__dll_func__ = None # object storing the compiled Fortran module
        self._make_f90_module()
        self._compile_func()
        #self._compile_jac()
    
    def _import(self):
        # import the dll    
        dll_func_name = self.__name__ # in case if BIND(C) option is enabled
        #dll_func_name = 'MODULE_{name}_mp_{name}'.format(name=self.__name__.upper())
        module = ct.CDLL(self.__module_f90_name__+'.dll')
        self.__dll_func__ = getattr(module,dll_func_name)        
        
    def _make_f90_module(self):
        """
        Prepare and save Fortran code for processing by Adifor
        and for importing by the main function.
        """
        # get parameter vector (including fixed ones)
        parvals = self.__params__.get_values()
        
        # split the code into lines 
        func_code = self.__code__
        func_code = textwrap.dedent(func_code)
        func_code = func_code.strip()
        func_code = func_code.split('\n')

        # update the name of the subroutine
        func_code[0] = re.sub('subroutine\s+(.+)\(','subroutine %s('%self.__name__,func_code[0])
        func_code[0] += ' BIND(C, NAME=\'%s\')'%self.__name__  # commenting this will possibly break the code portability

        # insert the export declaration
        func_code.insert(1,'!DEC$ ATTRIBUTES DLLEXPORT :: %s'%self.__name__)

        # get the list of model inputs
        parlist = re.search('subroutine\s+[^\(]+\(([^\(\)]+)\)',func_code[0]).group(1).split(',')
        pars = ','.join(parlist[2:])        
        
        # insert declaration for the interface parameter (assumed-shape arrays give compilation error)
        func_code.insert(2,'double precision, intent(out) :: %s'%parlist[0])
        func_code.insert(2,'double precision, intent(in) :: %s(np__),%s'%(parlist[1],','.join(parlist[2:])))
        
        # get the wrapper code for the function to be used from Fortran
        fortran_wrapper = textwrap.dedent("""
        subroutine wrapper_{name}(res,{pars})
                
        double precision,intent(out) :: res
        double precision,intent(in) :: {pars}
        
        call {name}(res,p,{pars})
        
        end subroutine
        """).format(name=self.__name__,pars=pars)
        
        # join the code lines back
        func_code = '\n'.join(func_code)
        
        # get the complete code of the module
        module_code = textwrap.dedent("""
        module {module_name}
        
        use iso_c_binding
        implicit none
        
        {data}
        {np__}
        {params}
        
        contains
        
        {function_lower_level}
        
        {function_higher_level}
        
        end module {module_name}
        """).format(module_name=self.__module_f90_name__,
            data=self.__data__.__get_code__(),
            np__=FortranValue('np__','integer',len(parvals)).__get_code__(),
            params=FortranArray('p','double precision',
                parvals).__get_code__(),
            function_lower_level=func_code,
            function_higher_level=fortran_wrapper)
        module_code = module_code.strip()
        
        # save the result
        with open(self.__module_f90_name__+'.f90','w') as f:
            f.write(module_code)
        
    def _compile_func(self):
        """
        Compile codes for function and Jacobian, and import the dll modules
        """
        # Compile and import function.
        compiler = IFort(files=[self.__module_f90_name__+'.f90'],options=['/dll','/traceback']) # debug
        #compiler = IFort(files=[self.__module_f90_name__+'.f90'],options=['/dll','/O3']) # release
        compiler.run()
        
    def _compile_jac(self):
        # Compile and import Jacobian.
        compiler = IFort(files=[self.__module_jac_f90_name__+'.f90'],options=['/dll','/O0','/g','/traceback']) # debug
        #compiler = IFort(files=[self.__module_jac_f90_name__+'.f90'],options=['/dll','/O3']) # release
        compiler.run()
          
    #@abstractmethod
    def __func__(self,params,*inputs):
        # make initial preparations
        if self.__dll_func__ is None:
            self._import()
        p = params.get_values()
        # setup the data
        input_pointers = [ct.pointer(ct.c_float(input)) for input in inputs]
        # call the function by passing the ctypes pointer using the numpy function:
        res = ct.c_double(); res_pointer = ct.pointer(res)
        self.__dll_func__(res_pointer,np.ctypeslib.as_ctypes(p),*input_pointers)
        return res.value

    def __jac__(self,params,*inputs):
        raise NotImplementedError()
            
            
# ===============================================================================
# =============================== /MODEL FORTRAN ================================
# ===============================================================================



from setuptools import find_packages, setup

setup(
    name="pysurfacefit",
    version="0.1",
    author="Roman Kochanov",
    author_email="",
    description="Multidimensional data fitting tool in Python",
    #url="",
    python_requires=">=3.7",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy",
        "numba",
        "tabulate",
        "scipy",
        "matplotlib",
        "jeanny3",
    ],
    entry_points = {
        'console_scripts': ['pysurfacefit=pysurfacefit.command_line:main']
    }
)
